/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux_DR2.1_Update1
 */
/* 
 * Copyright (c) 1988-1994, The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *	Utah $Hdr: grf.c 1.41 94/12/14$
 */

/*
 * Graphics display driver for HP 300/400/700/800 machines.
 * This is the hardware-independent portion of the driver.
 * Hardware access is through the machine dependent grf switch routines.
 */

#include "grf.h"
#if NGRF > 0

#include <types.h>
#include <sys/ioctl.h>
#include <device/tty.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <kern/spl.h>
#include <vm/vm_kern.h>
#include <vm/pmap.h>
#include <hp_pa/HP700/grfioctl.h>
#include <hp_pa/HP700/grfvar.h>
#include <hp_pa/HP700/itevar.h>

#include <machine/cpu.h>
#include <kern/misc_protos.h>

#include "ite.h"
#if NITE == 0
#define	iteon(u,f)
#define	iteoff(u,f)
#endif

struct grf_softc grf_softc[NGRF];

#ifdef MAPMEM
int grfexit();
struct mapmemops grfops = { (int (*)())0, (int (*)())0, grfexit, grfexit };
#ifdef HPUXCOMPAT
struct mapmemops grflckops = { (int (*)())0, (int (*)())0, grfexit, grfexit };
#endif
#endif

#ifdef DEBUG
int grfdebug = 0;
#define GDB_DEVNO	0x01
#define GDB_MMAP	0x02
#define GDB_IOMAP	0x04
#define GDB_LOCK	0x08
#endif

/* Forwards */
int grfopen(dev_t, int);
void grfclose(dev_t, int);
io_return_t grfgetstat(dev_t, int, int *, unsigned int *);
io_return_t grfsetstat(dev_t, int, int *, unsigned int);
int grfioctl(dev_t, int, caddr_t, int);
int grfselect(dev_t, int);
int grfmap(dev_t, int, int);
int grfon(dev_t);
int grfoff(dev_t);
int grfaddr(struct grf_softc *, int);
int grfmmap(dev_t, caddr_t*);
int grfunmmap(dev_t, caddr_t);
int grflock(struct grf_softc *,	int);
int grfunlock(struct grf_softc *);
int grfdevno(dev_t);
#ifdef MAPMEM
int grfmapin(struct mapmem *, int);
#endif
#ifdef HPUXCOMPAT
int hpuxgrfioctl(dev_t, int, caddr_t, int);
#endif

/*ARGSUSED*/
int 
grfopen(dev_t dev, int flags)
{
	int unit = GRFUNIT(dev);
	register struct grf_softc *gp = &grf_softc[unit];
	int error = 0;

	if (unit >= NGRF || (gp->g_flags & GF_ALIVE) == 0)
		return(ENXIO);
	if ((gp->g_flags & (GF_OPEN|GF_EXCLUDE)) == (GF_OPEN|GF_EXCLUDE))
		return(EBUSY);
#ifdef HPUXCOMPAT
	if (gp->g_flags & GF_BSDOPEN)
		return(EBUSY);
	else
		gp->g_flags |= GF_HPUXOPEN;
#endif
	/*
	 * First open.
	 * XXX: always put in graphics mode.
	 */
	error = 0;
	if ((gp->g_flags & GF_OPEN) == 0) {
		gp->g_flags |= GF_OPEN;
		error = grfon(dev);
	}
	return(error);
}

/*ARGSUSED*/
void
grfclose(dev_t dev, int flags)
{
	register struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];

	(void) grfoff(dev);
#ifdef HPUXCOMPAT
	(void) grfunlock(gp);
#endif
	gp->g_flags &= GF_ALIVE;
}

io_return_t
grfgetstat(
	dev_t		dev,
	int		flavor,
	int *		data,		/* pointer to OUT array */
	unsigned int	*count)
{
	int error;

	error = grfioctl(dev, flavor, (caddr_t)data, 0);
	*count = ((flavor >> 16) & 0x7F) / sizeof(int);
	return(error);
}

io_return_t
grfsetstat(
	dev_t		dev,
	int		flavor,
	int *		data,
	unsigned int	count)
{
	return(grfioctl(dev, flavor, (caddr_t)data, 0));
}

/*ARGSUSED*/
int
grfioctl(
	dev_t dev,
	int cmd,
	caddr_t data,
	int flag)
{
	register struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	int error;

#ifdef HPUXCOMPAT
	return(hpuxgrfioctl(dev, cmd, data, flag));
#endif
	error = 0;
	switch (cmd) {

	case GRFIOCGINFO:
		bcopy((caddr_t)&gp->g_display, data, sizeof(struct grfinfo));
		break;

	case GRFIOCON:
		error = grfon(dev);
		break;

	case GRFIOCOFF:
		error = grfoff(dev);
		break;

#ifdef MAPMEM
	case GRFIOCMAP:
		error = grfmmap(dev, (caddr_t *)data);
		break;

	case GRFIOCUNMAP:
		error = grfunmmap(dev, *(caddr_t *)data);
		break;
#endif

	default:
		error = EINVAL;
		break;

	}
	return(error);
}

/*ARGSUSED*/
int 
grfselect(dev_t dev, int rw)
{
	if (rw == D_READ)
		return(0);
	return(1);
}

/*ARGSUSED*/
int
grfmap(dev_t dev, int off, int prot)
{
	return(grfaddr(&grf_softc[GRFUNIT(dev)], off));
}

int
grfon(dev_t dev)
{
	int unit = GRFUNIT(dev);
	struct grf_softc *gp = &grf_softc[unit];

	/*
	 * XXX: iteoff call relies on devices being in same order
	 * as ITEs and the fact that iteoff only uses the minor part
	 * of the dev arg.
	 */
	iteoff(unit, 3);
	return((*gp->g_sw->gd_mode)(gp,
#ifdef HPOSFCOMPAT
				    GM_GRFON,
#else
				    (dev&GRFOVDEV) ? GM_GRFOVON : GM_GRFON,
#endif
				    (caddr_t)0));
}

int 
grfoff(dev_t dev)
{
	int unit = GRFUNIT(dev);
	struct grf_softc *gp = &grf_softc[unit];
	int error;

#ifdef MAPMEM
	(void) grfunmmap(dev, (caddr_t)0);
#endif
	error = (*gp->g_sw->gd_mode)(gp,
#ifdef HPOSFCOMPAT
				     GM_GRFOFF,
#else
				     (dev&GRFOVDEV) ? GM_GRFOVOFF : GM_GRFOFF,
#endif
				     (caddr_t)0);
	/* XXX: see comment for iteoff above */
	iteon(unit, 2);
	return(error);
}

int 
grfaddr(
	struct grf_softc *gp,
	int off)
{
#ifdef MAPMEM
	return(0);
#endif
	/* bogus */
	return(-1);
}

int 
grfmmap(dev_t dev, caddr_t *addrp)
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	int len;
#ifdef MAPMEM
	register struct mapmem *mp;
	int flags;
#endif

#ifdef DEBUG
	if (grfdebug & GDB_MMAP)
		printf("grfmmap(%d): addr %x\n", current_thread(), *addrp);
#endif
	*addrp = (caddr_t)((int)gp->g_display.gd_regaddr & 0xFC000000);
	len = 0x2000000;
#ifdef MAPMEM
	flags = MM_RW|MM_IO|MM_NOCORE;
	mp = mmalloc(minor(dev), addrp, len, flags, &grfops);
	if (mp == MMNIL)
		return(u.u_error);
	if (!mmmapin(mp, grfmapin)) {
		mmfree(mp);
		return(u.u_error);
	}
	if ((*gp->g_sw->gd_mode)(gp, GM_MAP, *addrp)) {
		mmmapout(mp);
		mmfree(mp);
		u.u_error = EINVAL;
		return(u.u_error);
	}
#endif
#ifndef IO_HACK
	if (!pmap_iomap((vm_offset_t)*addrp))
		return(EINVAL);
#endif
	return(0);
}

int
grfunmmap(dev_t dev, caddr_t addr)
{
#ifdef MAPMEM
	register struct mapmem *mp, **mpp;
	int found, unit = minor(dev);
#endif
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];

#ifdef DEBUG
	if (grfdebug & GDB_MMAP)
		printf("grfunmmap(%d): id %d addr %x\n",
		       current_thread(), minor(dev), addr);
#endif
#ifdef MAPMEM
	/* GCUNMAP has no argument on the PA */
	addr = 0;
	found = 0;
	mpp = &u.u_mmap;
	for (mp = *mpp; mp; mp = *mpp) {
		if (mp->mm_ops != &grfops || mp->mm_id != unit) {
			mpp = &mp->mm_next;
			continue;
		}
		if (addr &&
		    (addr < mp->mm_uva || addr >= mp->mm_uva+mp->mm_size)) {
			mpp = &mp->mm_next;
			continue;
		}
		grfexit(mp);
		found++;
	}
	if (found == 0)
		return(EINVAL);
#endif
	addr = (caddr_t)((int)gp->g_display.gd_regaddr & 0xFC000000);
#ifndef IO_HACK
	pmap_iounmap((vm_offset_t)addr);
#endif
	return(0);
}

/*
 * HP-UX compatibility routines
 */
#ifdef HPUXCOMPAT

/*ARGSUSED*/
int
hpuxgrfioctl(dev_t dev, int cmd, caddr_t data, int flag)
{
	register struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	int error;

	error = 0;
	switch (cmd) {

	case GCID:
		*(int *)data = gp->g_display.gd_id;
		break;

	case GCON:
		error = grfon(dev);
		break;

	case GCOFF:
		error = grfoff(dev);
		break;

	case GCLOCK:
	case GCLOCK_MINIMUM:
	case GCFASTLOCK:
		error = grflock(gp, 1);
		break;

	case GCUNLOCK:
	case GCUNLOCK_MINIMUM:
		error = grfunlock(gp);
		break;

	case GCAON:
	case GCAOFF:
		break;

	/* GCSTATIC is implied by our implementation */
	case GCSTATIC_CMAP:
	case GCVARIABLE_CMAP:
		break;

	/* map in control regs and frame buffer */
	case GCMAP:
	case GCMAP_HPUX: /* this should be cleaned up sometimes !! */
		error = grfmmap(dev, (caddr_t*)data);
		break;

	case GCUNMAP:
	case GCUNMAP_HPUX:
		error = grfunmmap(dev, *(caddr_t*)data);
#ifdef MAPMEM
		/* XXX: HP-UX uses GCUNMAP to get rid of GCSLOT memory */
		if (error)
			error = grflckunmmap(dev, *data);
#endif
		break;

#ifdef MAPMEM
	case GCSLOT:
	{
		struct gcslot_info *sp = (struct grf_slot *)data;

		sp->slot = grffindpid(gp);
		if (sp->slot)
			error = grflckmmap(dev, (caddr_t *)&sp->addr);
		else
			error = EINVAL;		/* XXX */
		break;
	}

#endif

	case GCDESCRIBE:
		error = (*gp->g_sw->gd_mode)(gp, GM_DESCRIBE, data);
		break;

#ifdef HPOSFCOMPAT
	case GCPHYSADDR:
	{
		struct grf_physaddr *pa = (struct grf_physaddr *)data;

		pa->fbaddr = (u_char*)gp->g_display.gd_fbaddr;
		pa->regaddr = (u_char*)gp->g_display.gd_regaddr;
		break;
	}
#endif

	default:
		error = EINVAL;
		break;
	}
	return(error);
}

int
grflock(
	struct grf_softc *gp,
	int block)
{
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grflock(%d): dev %x flags %x\n",
		       current_thread(), gp-grf_softc, gp->g_flags);
#endif
	/* XXX implement me XXX */

	return(0);
}

int 
grfunlock(struct grf_softc *gp)
{
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grfunlock(%d): dev %x flags %x\n",
		       current_thread(), gp-grf_softc, gp->g_flags);
#endif
	/* XXX implement me XXX */
	return(0);
}

#ifdef	HPOSFCOMPAT
#define bsdtohpuxdev(d)	((major(d) << 24) | minor(d))
#endif

/*
 * Convert a BSD style minor devno to HPUX style.
 * We cannot just create HPUX style nodes as they require 24 bits
 * of minor device number and we only have 8.
 * XXX: This may give the wrong result for remote stats of other
 * machines where device 10 exists.
 */
int
grfdevno(dev_t dev)
{
	int unit = GRFUNIT(dev);
	struct grf_softc *gp = &grf_softc[unit];
	int newdev;

	if (unit >= NGRF || (gp->g_flags&GF_ALIVE) == 0)
		return(bsdtohpuxdev(dev));
	newdev = (12 << 24) | 0x100000 | (unit << 8);

#ifdef DEBUG
	if (grfdebug & GDB_DEVNO)
		printf("grfdevno: dev %x newdev %x\n", dev, newdev);
#endif
	return(newdev);
}

#endif	/* HPUXCOMPAT */

/*
 * Mapped memory interface routines
 */
#ifdef MAPMEM

int 
grfmapin(
	struct mapmem *mp;
	int off)
{
	return(grfaddr(&grf_softc[GRFUNIT(mp->mm_id)], off));
}

grfexit(mp)
	struct mapmem *mp;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(mp->mm_id)];

#ifdef DEBUG
	if (grfdebug & GDB_MMAP)
		printf("grfexit(%d): id %d %x@%x\n",
		       current_thread(), mp->mm_id, mp->mm_size, mp->mm_uva);
#endif
#ifdef HPUXCOMPAT
	(void) grfunlock(gp);
	grfrmpid(gp);
#endif
	(void) (*gp->g_sw->gd_mode)(gp, GM_UNMAP, 0);
	mmmapout(mp);
	mmfree(mp);
}

#ifdef HPUXCOMPAT
/*
 * Processes involved in framebuffer mapping via GCSLOT are recorded in
 * an array of pids.  The first element is used to record the last slot used
 * (for faster lookups).  The remaining elements record up to GRFMAXLCK-1
 * process ids.  Returns a slot number between 1 and GRFMAXLCK or 0 if no
 * slot is available. 
 */
grffindpid(gp)
	struct grf_softc *gp;
{
	register short pid, *sp;
	register int i, limit;
	int ni;

	if (gp->g_pid == NULL) {
		gp->g_pid = (short *)
			malloc(GRFMAXLCK * sizeof(short), M_DEVBUF, M_WAITOK);
		bzero((caddr_t)gp->g_pid, GRFMAXLCK * sizeof(short));
	}
	pid = current_thread();
	ni = limit = gp->g_pid[0];
	for (i = 1, sp = &gp->g_pid[1]; i <= limit; i++, sp++) {
		if (*sp == pid)
			goto done;
		if (*sp == 0)
			ni = i;
	}
	i = ni;
	if (i < limit) {
		gp->g_pid[i] = pid;
		goto done;
	}
	if (++i == GRFMAXLCK)
		return(0);
	gp->g_pid[0] = i;
	gp->g_pid[i] = pid;
done:
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grffindpid(%d): slot %d of %d\n",
		       pid, i, gp->g_pid[0]);
#endif
	return(i);
}

grfrmpid(gp)
	struct grf_softc *gp;
{
	register short pid, *sp;
	register int limit, i;
	int mi;

	if (gp->g_pid == NULL || (limit = gp->g_pid[0]) == 0)
		return;
	pid = current_thread();
	limit = gp->g_pid[0];
	mi = 0;
	for (i = 1, sp = &gp->g_pid[1]; i <= limit; i++, sp++) {
		if (*sp == pid)
			*sp = 0;
		else if (*sp)
			mi = i;
	}
	i = mi;
	if (i < limit)
		gp->g_pid[0] = i;
#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grfrmpid(%d): slot %d of %d\n",
		       pid, sp-gp->g_pid, gp->g_pid[0]);
#endif
}

#define cialloc(sz)	malloc(sz, M_DEVBUF, M_WAITOK)

/*ARGSUSED*/
grflckmapin(mp, off)
	struct mapmem *mp;
{
	u_int pa = kvtophys((u_int)grf_softc[GRFUNIT(mp->mm_id)].g_lock);

#ifdef DEBUG
	if (grfdebug & GDB_LOCK)
		printf("grflckmapin(%d): va %x pa %x\n", current_thread(),
		       grf_softc[GRFUNIT(mp->mm_id)].g_lock, pa);
#endif
	return(pa >> PGSHIFT);
}

grflckmmap(dev, addrp)
	dev_t dev;
	caddr_t *addrp;
{
	struct grf_softc *gp = &grf_softc[GRFUNIT(dev)];
	register struct mapmem *mp;
	int grflckmapin();

#ifdef DEBUG
	if (grfdebug & (GDB_MMAP|GDB_LOCK))
		printf("grflckmmap(%d): addr %x\n", current_thread(), *addrp);
#endif
	if (gp->g_lock == NULL) {
		gp->g_lock = (struct grf_lockpage *) cialloc(NBPG);
		if (gp->g_lock == NULL)
			return(ENOMEM);
	}
	else
		/* XXX make sure kernel address is valid */
		mmvalidate((caddr_t)gp->g_lock, 1);

	mp = mmalloc(minor(dev), addrp, NBPG, MM_RW|MM_CI, &grflckops);
	if (mp == MMNIL)
		return(u.u_error);
	if (!mmmapin(mp, grflckmapin)) {
		mmfree(mp);
		return(u.u_error);
	}
	return(0);
}

grflckunmmap(dev, addr)
	dev_t dev;
	caddr_t addr;
{
	register struct mapmem *mp;
	int unit = minor(dev);

#ifdef DEBUG
	if (grfdebug & (GDB_MMAP|GDB_LOCK))
		printf("grflckunmmap(%d): id %d addr %x\n",
		       current_thread(), unit, addr);
#endif
	for (mp = u.u_mmap; mp; mp = mp->mm_next)
		if (mp->mm_ops == &grflckops && mp->mm_id == unit &&
		    mp->mm_uva == addr) {
			grfexit(mp);
			return(0);
		}
	return(EINVAL);
}

#endif	/* HPUXCOMPAT */
#endif	/* MAPMEM */
#endif	/* NGRF > 0 */
