/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 * 
 */
/*
 * Copyright 1996 1995 by Apple Computer, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * APPLE COMPUTER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL APPLE COMPUTER BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux_DR2.1_Update1
 */

/* PMach Operating System
 * Copyright (c) 1995 Santa Clara University
 * All Rights Reserved.
 */
/* Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS AS-IS
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */

#include <asc.h>

#if NASC > 0

#include <mach_debug.h>

#include <mach/std_types.h>
#include <types.h>
#include <chips/busses.h>
#include <scsi/compat_30.h>
#include <machine/spl.h>

#include <scsi/scsi.h>
#include <scsi/scsi2.h>

#include <scsi/adapters/scsi_53C94.h>
#include <scsi/scsi_defs.h>
#include <scsi/adapters/scsi_dma.h>
#include <ppc/proc_reg.h>
#include <ppc/powermac.h>
#include <ppc/pmap.h> /* For kvtophys */

#include <scsi.h>

typedef struct dma_softc {
	volatile unsigned char	*base;
	volatile unsigned char	*ctrl;
	volatile unsigned char	*cur_ptr;
} dma_softc_t;

dma_softc_t	dma_softc[NASC] = {
{ 	(unsigned char *) SCSI_DMA_BASE ,	/* Channel A base */
	(unsigned char *) SCSI_DMA_BASE+0x8,	/* Channel A status */
	(unsigned char *) SCSI_DMA_BASE+0x10	/* Channel A ptr */
}
#if NASC > 1
,{
	/* Note, the second DMA channels aren't exactly aligned
	 * correctly. (the second SCSI dma channel is more
	 * of an after-thought.) Hence, the reasons for the
	 * funnny offsets
	 */

 	(unsigned char *) SCSI_DMA_BASE+0x4,	/* Channel B base */
	(unsigned char *) SCSI_DMA_BASE+0x9,	/* Channel B status */
	(unsigned char *) SCSI_DMA_BASE+0x14	/* Channel B ptr */
}
#endif
};

/* Note - there are two DMA channels available, we hardcode access to
 * just one of them, this could be tidied up. Currently PDM machines
 * only use the first channel
 */

#if MACH_DEBUG

int amic_dma_debug=0;	/* Change to 1 or 2 to enable debugging traces */

#define DPRINTF(x) if (amic_dma_debug) { printf("%s : ",__FUNCTION__); printf x; }
#else
#define DPRINTF(x)
#endif /* MACH_DEBUG */

/* Prototypes for functions here */
void		amic_dma_base_put(dma_softc_t *dma, unsigned char *base_addr);
opaque_t	amic_dma_init(
			      int dev_unit,
			      vm_offset_t base,
			      int *dma_bsizep);
void		amic_dma_new_target(
			      dma_softc_t *dma_state,
			      target_info_t *tgt);
void		amic_dma_map(
			     dma_softc_t *dma_state,
			     target_info_t *tgt);
int		amic_dma_start_cmd(
				   dma_softc_t *dma_state,
				   target_info_t *tgt);
void		amic_dma_end_xfer(
				  dma_softc_t *dma_state,
				  target_info_t *tgt,
				  int xferred);
void		amic_dma_end_cmd(
				 dma_softc_t *dma_state,
				 target_info_t *tgt,
				 io_req_t ior);
int		amic_dma_start_datain(
				      dma_softc_t *dma_state,
				      target_info_t *tgt);
int		amic_dma_start_msgin(
				     dma_softc_t *dma_state,
				     target_info_t *tgt);
void		amic_dma_end_msgin(
				   dma_softc_t *dma_state,
				   target_info_t *tgt);
boolean_t	amic_dma_start_dataout(
				       dma_softc_t *dma_state,
				       target_info_t *tgt,
				       unsigned *regp,
				       unsigned value);
int		amic_dma_restart_datain_1(
				  dma_softc_t *dma_state,
				  target_info_t *tgt);
int		amic_dma_restart_datain_2(
				  dma_softc_t *dma_state,
				  target_info_t *tgt,
				  int xferred);
void		amic_dma_restart_datain_3(
				  dma_softc_t *dma_state,
				  target_info_t *tgt);
int		amic_dma_restart_dataout_1(
				   dma_softc_t *dma_state,
				   target_info_t *tgt);
int		amic_dma_restart_dataout_2(
				   dma_softc_t *dma_state,
				   target_info_t *tgt,
				   int xferred);
int		amic_dma_restart_dataout_3(
				   dma_softc_t *dma_state,
				   target_info_t *tgt,
				   unsigned *regp);
void		amic_dma_restart_dataout_4(
				   dma_softc_t *dma_state,
				   target_info_t *tgt);
boolean_t	amic_dma_disconn_1(
				   dma_softc_t *dma_state,
				   target_info_t *tgt,
				   int xferred);
boolean_t	amic_dma_disconn_2(
				   dma_softc_t *dma_state,
				   target_info_t *tgt);
boolean_t	amic_dma_disconn_3(
				   dma_softc_t *dma_state,
				   target_info_t *tgt,
				   int xferred);
boolean_t	amic_dma_disconn_4(
				   dma_softc_t *dma_state,
				   target_info_t *tgt,
				   int xferred);
boolean_t	amic_dma_disconn_5(
				   dma_softc_t *dma_state,
				   target_info_t *tgt,
				   int xferred);
void		amic_dma_disconn_callback(
					  dma_softc_t *dma_state,
					  target_info_t *tgt);

#define	u_min(a, b)	(((a) < (b)) ? (a) : (b))


unsigned long
amic_dma_count(unsigned char *base_address, unsigned long count)
{
	unsigned long page	= (unsigned long) base_address & 0xfff;

	if (page+count > 0x1000) {
		return	0x1000 - page;
	} else
		return count;
}

void amic_dma_base_put(dma_softc_t *dmap, unsigned char *base_addr)
{
	vm_offset_t base_val = kvtophys((vm_offset_t)base_addr);

	assert(base_val != 0);
	
	DPRINTF(("addr = 0x%08x, phys addr=0x%08x\n",base_addr,base_val));
	
	assert(dmap != (dma_softc_t *)NULL);

	/* amic dma apparently requires 8 byte alignment for start addr */
	assert((base_val & 7) == 0);
	
	dmap->base[0] = (unsigned char)((base_val >> 24) & 0xff);
	dmap->base[1] = (unsigned char)((base_val >> 16) & 0xff);
	dmap->base[2] = (unsigned char)((base_val >>  8) & 0xff);
	dmap->base[3] = (unsigned char)((base_val) & 0xff);

	eieio();
}


opaque_t amic_dma_init(int dev_unit, vm_offset_t base, int *dma_bsizep)
{
	DPRINTF(("dev_unit = %d, base=0x%08x\n",dev_unit,base));
	
	*dma_bsizep = 20;

	/* We assume that the dma engine is present */

	return (opaque_t) &dma_softc[dev_unit];
}

/*
 * Statically allocated command & temp buffers
 * This way we can attach/detach drives on-fly
 */
#define	PER_TGT_BUFF_DATA	256

/*
 * The +4's are to ensure that we assign a double-word aligned offset,
 * this structure is only single-word aligned
 */

static char	amic_buffer[NASC * 8 * PER_TGT_BUFF_DATA + 4];

void amic_dma_new_target(dma_softc_t *dma_state, target_info_t *tgt)
{
	DPRINTF(("called with dma_state = 0x%08x, target id=0x%x",
		 dma_state,tgt->target_id));

	if (dma_state) {
		tgt->cmd_ptr = &amic_buffer[(((tgt->unit_no+1)*tgt->target_id) *
					    PER_TGT_BUFF_DATA) + 4];
		tgt->cmd_ptr = (char*)  ((vm_offset_t) tgt->cmd_ptr & ~7);
		tgt->dma_ptr = 0;
	}
	return;
}

void	amic_dma_map(dma_softc_t *dmap, target_info_t *tgt)
{
	DPRINTF(("called with command tgt->cur_cmd=0x%x\n",tgt->cur_cmd));

	/*
	 * We cannot do real DMA
	 */

	if ((tgt->cur_cmd == SCSI_CMD_WRITE) ||
	    (tgt->cur_cmd == SCSI_CMD_LONG_WRITE)) {
		io_req_t	ior = tgt->ior;
		unsigned int	max_dma;
		register int	count;

		max_dma = 4096; /* tgt->max_dma_data; */
		count = ior->io_count;
		tgt->transient_state.out_count = u_min(count, max_dma);
	} else {
		tgt->transient_state.out_count  = 0;
	}

	tgt->transient_state.dma_offset = 0;
	tgt->dma_ptr = 0;
	return ;
}

void amic_dma_stop(dma_softc_t *dmap)
{
	/*
	 * Flush out the DMA, and kill it.
	 */

	eieio();
	if ((*dmap->ctrl & MASK8(SCSI_DMA_CTRL_DIR)) == 0) {
		*dmap->ctrl |= MASK8(SCSI_DMA_CTRL_FLUSH);
		do {
			eieio();
		} while (*dmap->ctrl & MASK8(SCSI_DMA_CTRL_FLUSH));
	}

	*dmap->ctrl &= ~MASK8(SCSI_DMA_CTRL_RUN); eieio();

	while (*dmap->ctrl & MASK8(SCSI_DMA_CTRL_RUN))
		eieio();

}

void amic_dma_reset(dma_softc_t *dmap)
{
	*dmap->ctrl = MASK8(SCSI_DMA_CTRL_RESET); eieio();

	while (*dmap->ctrl & MASK8(SCSI_DMA_CTRL_RUN))
		eieio();
}

int	amic_dma_start_cmd(dma_softc_t *dmap, target_info_t *tgt)
{
	/* Taken from SCU code - start_cmd isn't used */
	return 0;
}

void 	amic_dma_end_xfer(dma_softc_t *dmap, target_info_t *tgt, int xferred)
{
	amic_dma_stop(dmap);

	tgt->transient_state.copy_count = 0;
	return;
}


void	amic_dma_end_cmd(dma_softc_t *dmap, target_info_t *tgt, io_req_t ior)
{
	tgt->dma_ptr = 0;
	tgt->transient_state.dma_offset = 0;

	return;
}


int	amic_dma_start_datain(dma_softc_t *dmap, target_info_t *tgt)
{
	int  data_in_count;
	unsigned char *ch_temp;
	char		*addr;
	
	amic_dma_reset(dmap);

	ch_temp = tgt->cmd_ptr;
	
	DPRINTF(("tgt->cmd_ptr address = %08x\n",tgt->cmd_ptr));
	
	
	if (tgt->cur_cmd == SCSI_CMD_READ
	|| tgt->cur_cmd == SCSI_CMD_LONG_READ)
		addr = tgt->ior->io_data;
	else
		addr = tgt->cmd_ptr;


	invalidate_cache_v((vm_offset_t) addr, tgt->transient_state.in_count);

	data_in_count = amic_dma_count((unsigned char *) addr, tgt->transient_state.in_count);
	tgt->transient_state.dma_offset = 0;
	tgt->dma_ptr = addr;

	amic_dma_base_put(dmap, (unsigned char *) addr);

	*dmap->ctrl = MASK8(SCSI_DMA_CTRL_RUN) | MASK8(SCSI_DMA_CTRL_MODE1); eieio();

	DPRINTF(("data_in_count = %d\n",data_in_count));
	
	return data_in_count;
}

int	amic_dma_start_msgin(dma_softc_t *dmap,  target_info_t *tgt)
{
	return 0;
}

void	amic_dma_end_msgin(dma_softc_t *dmap, target_info_t *tgt)
{
	return;
}


boolean_t amic_dma_start_dataout(
			      dma_softc_t *dmap,
			      target_info_t	*tgt,
			      unsigned *regp,
			      unsigned	value)
{
	int  data_out_count;
	unsigned char *ch_temp,*ch_temp_1;
	char	*addr;
	int i;
	
	amic_dma_reset(dmap);
	
	DPRINTF(("tgt->cmd_ptr address = %08x\n",tgt->cmd_ptr));
	
	
	if (tgt->cur_cmd == SCSI_CMD_MODE_SELECT) {
		scsi_cmd_mode_select_t *cmd_temp;
		cmd_temp = (scsi_cmd_mode_select_t*)tgt->cmd_ptr;
		ch_temp =(unsigned char*)(cmd_temp+1);
		ch_temp_1 = (unsigned char*)tgt->cmd_ptr;
		for (i=0; i < tgt->transient_state.out_count; i++)
			*ch_temp_1++ = *ch_temp++;

		addr = tgt->cmd_ptr;
	} else if (tgt->cur_cmd == SCSI_CMD_WRITE
		|| tgt->cur_cmd == SCSI_CMD_LONG_WRITE) 
		addr = tgt->ior->io_data;
	else
		addr = tgt->cmd_ptr;

	data_out_count = amic_dma_count((unsigned char *) addr, tgt->transient_state.out_count);
	DPRINTF(("data_out_count = %d\n",data_out_count));

	tgt->transient_state.dma_offset = 0;
	tgt->dma_ptr = addr;

	flush_cache_v((vm_offset_t) addr, tgt->transient_state.out_count);

	amic_dma_base_put(dmap, addr);
	
	DPRINTF(("tgt->cmd_ptr address = %08x\n",tgt->cmd_ptr));
	*dmap->ctrl = MASK8(SCSI_DMA_CTRL_DIR) |
		    MASK8(SCSI_DMA_CTRL_RUN) | MASK8(SCSI_DMA_CTRL_MODE1); eieio();
	/*
 	 * This is a hack until the SCSI driver gets
	 * rewritten. 4-10-96 Michael (MEB)
	 */
	tgt->transient_state.copy_count = data_out_count;
	
	return TRUE;
}

int amic_dma_restart_datain_1(dma_softc_t *dmap, target_info_t *tgt)
{
	unsigned char 	*addr;
	int		in_count;

	if (tgt->dma_ptr == 0)
		return	amic_dma_start_datain(dmap, tgt);

	addr = (unsigned char *) tgt->dma_ptr + tgt->transient_state.dma_offset;

	in_count = amic_dma_count(addr, tgt->transient_state.in_count);
	amic_dma_reset(dmap);

	amic_dma_base_put(dmap, addr);

	*dmap->ctrl = MASK8(SCSI_DMA_CTRL_RUN) | MASK8(SCSI_DMA_CTRL_MODE1); eieio();
	
	return in_count;
}

int amic_dma_restart_datain_2(dma_softc_t *dmap,
			      target_info_t *tgt,
			      int xferred)
{
	assert((xferred & 0x7) == 0);
	assert(xferred >= 0);
	tgt->transient_state.dma_offset += xferred;
	return amic_dma_restart_datain_1(dmap , tgt);
}

void amic_dma_restart_datain_3(dma_softc_t *dma_state, target_info_t *tgt)
{
	DPRINTF(("dma_ops_restart_datain_3\n"));
	return ;
}

int amic_dma_restart_dataout_1(dma_softc_t *dmap, target_info_t *tgt)
{
	int		out_count;
	unsigned char	*addr;

	if (tgt->dma_ptr == 0) {
		amic_dma_start_dataout(dmap, tgt, 0, 0);
		return	tgt->transient_state.copy_count;
	}

	addr = (unsigned char *) tgt->dma_ptr + tgt->transient_state.dma_offset;
	out_count = amic_dma_count(addr, tgt->transient_state.out_count);
	
	return out_count;
}

int amic_dma_restart_dataout_2(dma_softc_t *dmap,
			       target_info_t *tgt,
			       int xferred)
{
	assert(xferred >= 0);
	assert((xferred & 0x7) == 0);
	tgt->transient_state.dma_offset += xferred;

	return amic_dma_restart_dataout_1(dmap, tgt);
}

int amic_dma_restart_dataout_3(dma_softc_t *dmap,
			       target_info_t *tgt,
			       unsigned *regp)
{
	unsigned char	*addr;

	addr = (unsigned char *) tgt->dma_ptr + tgt->transient_state.dma_offset;

	amic_dma_reset(dmap);
	amic_dma_base_put(dmap, addr);

	*dmap->ctrl = MASK8(SCSI_DMA_CTRL_DIR) |
		    MASK8(SCSI_DMA_CTRL_RUN) | MASK8(SCSI_DMA_CTRL_MODE1); eieio();
	
	return 0;
}

void amic_dma_restart_dataout_4(dma_softc_t *dma_state, target_info_t *tgt)
{
	DPRINTF(("dma_ops_restart_dataout_4\n"));
	return ;
}

boolean_t amic_dma_disconn_1(dma_softc_t	*dmap,
			     target_info_t 	*tgt,
			     int 		xferred)
{
	assert(xferred >= 0);
	assert((xferred & 0x7) == 0);

	amic_dma_stop(dmap);
	tgt->transient_state.dma_offset += xferred;
	tgt->transient_state.copy_count += xferred;
	assert(xferred >= 0);
	return 0;
}

boolean_t amic_dma_disconn_2(dma_softc_t *dmap, target_info_t *tgt)
{
	amic_dma_stop(dmap);
	return 0;
}

#define	amic_dma_disconn_3		amic_dma_disconn_1
#define	amic_dma_disconn_4		amic_dma_disconn_1
#define	amic_dma_disconn_5		amic_dma_disconn_1
#define	amic_dma_disconn_callback	0	/* Never called */

scsi_dma_ops_t amic_dma_ops = {
#if 0
	amic_dma_init,
	amic_dma_new_target,
	amic_dma_map,
	amic_dma_start_cmd,
	amic_dma_end_xfer,
	amic_dma_end_cmd,
	amic_dma_start_datain,
	amic_dma_start_msgin,
	amic_dma_end_msgin,
	amic_dma_start_dataout,
	amic_dma_restart_datain_1,
	amic_dma_restart_datain_2,
	amic_dma_restart_datain_3,
	amic_dma_restart_dataout_1,
	amic_dma_restart_dataout_2,
	amic_dma_restart_dataout_3,
	amic_dma_restart_dataout_4,
	amic_dma_disconn_1,
	amic_dma_disconn_2,
	amic_dma_disconn_3,
	amic_dma_disconn_4,
	amic_dma_disconn_5,
	amic_dma_disconn_callback
#else
(opaque_t (*)(int, vm_offset_t, int *)) 	amic_dma_init,
(void (*)( opaque_t, target_info_t *)) 		amic_dma_new_target,
(void (*)( opaque_t , target_info_t *)) 	amic_dma_map,
(int (*)(opaque_t , target_info_t *)) 		amic_dma_start_cmd,
(void (*)(opaque_t, target_info_t *, int)) 	amic_dma_end_xfer,
(void (*)( opaque_t , target_info_t *,io_req_t )) 	amic_dma_end_cmd,
(int (*)( opaque_t , target_info_t *)) 		amic_dma_start_datain,
(int (*)( opaque_t , target_info_t *)) 		amic_dma_start_msgin,
(void (*)( opaque_t , target_info_t *)) 	amic_dma_end_msgin,
(boolean_t (*)( opaque_t , target_info_t *, unsigned *, unsigned)) 	amic_dma_start_dataout,
(int (*)( opaque_t , target_info_t *)) 	amic_dma_restart_datain_1,
(int (*)( opaque_t , target_info_t *, int )) 	amic_dma_restart_datain_2,
( void (*)( opaque_t , target_info_t *)) 	amic_dma_restart_datain_3,
(int (*)( opaque_t , target_info_t *)) 	amic_dma_restart_dataout_1,
(int (*)( opaque_t , target_info_t *, int )) 	amic_dma_restart_dataout_2,
(int (*)( opaque_t , target_info_t *, unsigned *regp)) 	amic_dma_restart_dataout_3,
(void (*)( opaque_t , target_info_t *)) 	amic_dma_restart_dataout_4,
(boolean_t (*)( opaque_t , target_info_t *, int )) 	amic_dma_disconn_1,
(boolean_t (*)( opaque_t , target_info_t *)) 	amic_dma_disconn_2,
(boolean_t (*)( opaque_t , target_info_t *, int )) 	amic_dma_disconn_3,
(boolean_t (*)( opaque_t , target_info_t *, int )) 	amic_dma_disconn_4,
(boolean_t (*)( opaque_t , target_info_t *, int )) 	amic_dma_disconn_5,
(void (*)( opaque_t , target_info_t *)) 	amic_dma_disconn_callback
#endif
};

/* Initialise the SCSI subsystem to use this dma */

void amic_init(void)
{
	asc_set_dmaops(0, &amic_dma_ops);

#if NASC > 1
	asc_set_dmaops(1, &amic_dma_ops);
#endif /* NASC > 1 */
}

#endif /* NASC > 0 */
