/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux_DR2.1_Update1
 */
/*
 * Mach Operating System
 * Copyright (c) 1990,1991,1992 The University of Utah and
 * the Center for Software Science (CSS).
 * Copyright (c) 1991,1987 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation,
 * and that all advertising materials mentioning features or use of
 * this software display the following acknowledgement: ``This product
 * includes software developed by the Center for Software Science at
 * the University of Utah.''
 *
 * CARNEGIE MELLON, THE UNIVERSITY OF UTAH AND CSS ALLOW FREE USE OF
 * THIS SOFTWARE IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY
 * OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF
 * THIS SOFTWARE.
 *
 * CSS requests users of this software to return to css-dist@cs.utah.edu any
 * improvements that they make and grant CSS redistribution rights.
 *
 * Carnegie Mellon requests users of this software to return to
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 *
 * 	Utah $Hdr: pmap.c 1.28 92/06/23$
 *	Author: Mike Hibler, Bob Wheeler, University of Utah CSS, 10/90
 */
 
/*
 *	Manages physical address maps for powerpc.
 *
 *	In addition to hardware address maps, this
 *	module is called upon to provide software-use-only
 *	maps which may or may not be stored in the same
 *	form as hardware maps.  These pseudo-maps are
 *	used to store intermediate results from copy
 *	operations to and from address spaces.
 *
 *	Since the information managed by this module is
 *	also stored by the logical address mapping module,
 *	this module may throw away valid virtual-to-physical
 *	mappings at almost any time.  However, invalidations
 *	of virtual-to-physical mappings must be done as
 *	requested.
 *
 *	In order to cope with hardware architectures which
 *	make virtual-to-physical map invalidates expensive,
 *	this module may delay invalidate or reduced protection
 *	operations until such time as they are actually
 *	necessary.  This module is given full information to
 *	when physical maps must be made correct.
 *	
 */

/*
 * CAVAETS:
 *
 *	Needs more work for MP support
 */

#include <debug.h>
#include <mach_vm_debug.h>

#include <kern/thread.h>
#include <mach/vm_attributes.h>
#include <mach/vm_param.h>
#include <kern/spl.h>

#include <kern/misc_protos.h>
#include <ppc/misc_protos.h>
#include <ppc/gdb_defs.h> /* for kgdb_kernel_in_pmap used in assert */
#include <ppc/proc_reg.h>
#include <ppc/mem.h>
#include <ppc/pmap.h>
#include <ppc/pmap_internals.h>
#include <ppc/new_screen.h>
#include <ppc/POWERMAC/powermac.h>
#include <ppc/POWERMAC/video_board.h>
#include <ppc/POWERMAC/video_pdm.h>

extern unsigned int	avail_remaining;

vm_offset_t		avail_next;
int          		current_free_region;	/* Used in pmap_next_page */

/* forward */
void pmap_activate(pmap_t pmap, thread_t th, int which_cpu);
void pmap_deactivate(pmap_t pmap, thread_t th, int which_cpu);
void copy_to_phys(vm_offset_t sva, vm_offset_t dpa, int bytecount);
#if MACH_VM_DEBUG
int pmap_list_resident_pages(pmap_t pmap, vm_offset_t *listp, int space);
#endif

#if DEBUG
#define PDB_USER	0x01	/* exported functions */
#define PDB_MAPPING	0x02	/* low-level mapping routines */
#define PDB_ENTER	0x04	/* pmap_enter specifics */
#define PDB_COPY	0x08	/* copy page debugging */
#define PDB_ZERO	0x10	/* zero page debugging */
#define PDB_WIRED	0x20	/* things concerning wired entries */
#define PDB_PTEG	0x40	/* PTEG overflows */
#define PDB_MASSIVE	0x80	/* Massive costly assert checks */
int pmdebug = PDB_PTEG;
#endif

struct pmap	kernel_pmap_store;
pmap_t		kernel_pmap;
struct zone	*pmap_zone;		/* zone of pmap structures */
boolean_t	pmap_initialized = FALSE;

/*
 * Virtual-to-physical translations are done by looking the address up
 * in the machine hash tables, searching the overflow PTEs if necessary.
 */
queue_head_t *vtop_table;

/*
 * Physical-to-virtual translations are handled by inverted page table
 * structures, phys_tables.  Multiple mappings of a single page are handled
 * by linking the affected mapping structures. We initialise one region
 * for phys_tables of the physical memory we know about, but more may be
 * added as it is discovered (eg. by drivers).
 */
struct phys_entry *phys_table;	/* For debugging */

/*
 * XXX use mpqueue_head_t's??
 * (no one else does, are they coming or going?)
 */
queue_head_t		free_mapping;	   /* list of free mapping structs */
decl_simple_lock_data(,free_mapping_lock) /* and lock */

pmap_t			free_pmap;	   /* list of free pmaps */
decl_simple_lock_data(,free_pmap_lock)   /* and lock */

decl_simple_lock_data(,pmap_lock)	   /* XXX this is all broken */

struct mapping *firstmap, *lastmap;	/* XXX for post-mortems */

unsigned	prot_bits[8];

int		sid_counter=0;	/* seed for SID generator */

/*
 * Function to get index into phys_table for a given physical address
 */

struct phys_entry *pmap_find_physentry(vm_offset_t pa)
{
	int i;
	struct phys_entry *entry;

	for (i = mem_regions_count-1; i >= 0; i--) {
		if (pa < mem_regions[i].start)
			continue;
		if (pa >= mem_regions[i].end)
			return PHYS_NULL;
		
		entry = &mem_regions[i].phys_table[(pa - mem_regions[i].start) >> PPC_PGSHIFT];
		return entry;
	}
	printf("NMGS DEBUG : pmap_find_physentry 0x%08x out of range\n",pa);
	return PHYS_NULL;
}

/*
 * kern_return_t
 * pmap_add_physical_memory(vm_offset_t spa, vm_offset_t epa,
 *                          boolean_t available, unsigned int attr)
 *	Allocate some extra physentries for the physical addresses given,
 *	specifying some default attribute that on the powerpc specifies
 *      the default cachability for any mappings using these addresses
 *	If the memory is marked as available, it is added to the general
 *	VM pool, otherwise it is not (it is reserved for card IO etc).
 */


kern_return_t pmap_add_physical_memory(vm_offset_t spa, vm_offset_t epa,
				       boolean_t available, unsigned int attr)
{
	int i,j;
	spl_t s;
	struct phys_entry *phys_table;

	/* Only map whole pages */

	spa = trunc_page(spa);
	epa = round_page(epa);

	/* First check that the region doesn't already exist */

	assert (epa >= spa);
	for (i = 0; i < mem_regions_count; i++) {
		/* If we're below the next region, then no conflict */
		if (epa < mem_regions[i].start)
			break;
		if (spa < mem_regions[i].end) {
#if DEBUG
			printf("pmap_add_physical_memory(0x%08x,0x%08x,0x%08x) - memory already present\n",spa,epa,attr);
#endif /* DEBUG */
			return KERN_NO_SPACE;
		}
	}

	/* Check that we've got enough space for another region */
	if (mem_regions_count == MEM_REGION_MAX)
		return KERN_RESOURCE_SHORTAGE;

	/* Once here, i points to the mem_region above ours in physical mem */

	/* allocate a new phys_table for this new region */

	phys_table =  (struct phys_entry *)
		kalloc(sizeof(struct phys_entry) * atop(epa-spa));

	/* Initialise the new phys_table entries */
	for (j = 0; j < atop(epa-spa); j++) {
		queue_init(&phys_table[j].phys_link);
		/* We currently only support these two attributes */
		assert((attr == PTE_WIMG_CACHED) ||
		       (attr == PTE_WIMG_UNCACHED));
		phys_table[j].pte1.bits.wimg = attr;
	}
	s = splvm();
	
	/* Move all the phys_table entries up some to make room in
	 * the ordered list.
	 */
	for (j = mem_regions_count; j > i ; j--)
		mem_regions[j] = mem_regions[j-1];

	/* Insert a new entry with some memory to back it */

	mem_regions[i].start 	     = spa;
	mem_regions[i].end           = epa;
	mem_regions[i].phys_table    = phys_table;

	mem_regions_count++;
	splx(s);

	if (available) {
		printf("warning : pmap_add_physical_mem() "
		       "available not yet supported\n");
	}

	return KERN_SUCCESS;
}

/*
 * ppc_protection_init()
 *	Initialise the user/kern_prot_codes[] arrays which are 
 *	used to translate machine independent protection codes
 *	to powerpc protection codes. The PowerPc can only provide
 *      [no rights, read-only, read-write]. Read implies execute.
 * See PowerPC 601 User's Manual Table 6.9
 */
void
ppc_protection_init(void)
{
	prot_bits[VM_PROT_NONE | VM_PROT_NONE  | VM_PROT_NONE]     = 0;
	prot_bits[VM_PROT_READ | VM_PROT_NONE  | VM_PROT_NONE]     = 3;
	prot_bits[VM_PROT_READ | VM_PROT_NONE  | VM_PROT_EXECUTE]  = 3;
	prot_bits[VM_PROT_NONE | VM_PROT_NONE  | VM_PROT_EXECUTE]  = 3;
	prot_bits[VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_NONE]     = 2;
	prot_bits[VM_PROT_READ | VM_PROT_WRITE | VM_PROT_NONE]     = 2;
	prot_bits[VM_PROT_NONE | VM_PROT_WRITE | VM_PROT_EXECUTE]  = 2;
	prot_bits[VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE]  = 2;
}

/*
 * pmap_map(va, spa, epa, prot)
 *	is called during boot to map memory in the kernel's address map.
 *	A virtual address range starting at "va" is mapped to the physical
 *	address range "spa" to "epa" with machine independent protection
 *	"prot".
 *
 *	"va", "spa", and "epa" are byte addresses and must be on machine
 *	independent page boundaries.
 */
vm_offset_t
pmap_map(
	vm_offset_t va,
	vm_offset_t spa,
	vm_offset_t epa,
	vm_prot_t prot)
{

	if (spa == epa)
		return(va);

	assert(epa > spa);

	while (spa < epa) {
		pmap_enter(kernel_pmap, va, spa, prot, TRUE);

		va += PAGE_SIZE;
		spa += PAGE_SIZE;
	}
	return(va);
}

/*
 * pmap_map_bd(va, spa, epa, prot)
 *	Back-door routine for mapping kernel VM at initialisation.
 *	Used for mapping memory outside the known physical memory
 *      space, with caching disabled. Designed for use by device probes.
 * 
 *	A virtual address range starting at "va" is mapped to the physical
 *	address range "spa" to "epa" with machine independent protection
 *	"prot".
 *
 *	"va", "spa", and "epa" are byte addresses and must be on machine
 *	independent page boundaries.
 *
 * WARNING: The current version of memcpy() can use the dcbz instruction
 * on the destination addresses.  This will cause an alignment exception
 * and consequent overhead if the destination is caching-disabled.  So
 * avoid memcpy()ing into the memory mapped by this function.
 *
 * also, many other pmap_ routines will misbehave if you try and change
 * protections or remove these mappings, they are designed to be permanent.
 */
vm_offset_t
pmap_map_bd(
	vm_offset_t va,
	vm_offset_t spa,
	vm_offset_t epa,
	vm_prot_t prot)
{
	pte_t  *pte;
	struct phys_entry* pp;
	struct mapping *mp;

	if (spa == epa)
		return(va);

	assert(epa > spa);

	while (spa < epa) {
		
		pte = find_or_allocate_pte(PPC_SID_KERNEL, va, TRUE);

		assert(pte != PTE_NULL);

		/* remapping of already-mapped addresses is OK, we just
		 * trample on the PTE, but make sure we're mapped to same
		 * address
		 */
		if (pte->pte0.bits.valid)
			assert(trunc_page(pte->pte1.word) == spa);

		invalidate_tlb(PPC_SID_KERNEL, va);

		/* Set up pte -
		 *   pte0 already contains everything except the valid bit.
		 */

		pte->pte1.bits.protection = prot_bits[prot];
		pte->pte1.bits.phys_page = spa >> PPC_PGSHIFT;
		pte->pte1.bits.wimg = PTE_WIMG_UNCACHED;

		/* make sure those pte1 writes above have completed */
		isync();
		pte->pte0.bits.valid = TRUE;	/* Validate pte entry */

		WRITEBACK_CACHED_PTE(pte);

		/* If memory is mappable (is in phys table) then set
		 * default cach attributes to non-cached and verify
		 * that there aren't any non-cached mappings which would
		 * break the processor spec (and hang the machine).
		 */
		pp = pmap_find_physentry(spa);
		if (pp != PHYS_NULL) {
			/* Set the default mapping type */
			pp->pte1.bits.wimg = PTE_WIMG_UNCACHED;
#if DEBUG
			/* Belt and braces check on mappings */
			queue_iterate(&pp->phys_link, mp,
				      struct mapping *, phys_link) {
				assert(mp->pte->pte1.bits.wimg ==
				       PTE_WIMG_UNCACHED);
			}
#endif /* DEBUG */
		}
		va += PAGE_SIZE;
		spa += PAGE_SIZE;
	}
	return(va);
}

/*
 *	Bootstrap the system enough to run with virtual memory.
 *	Map the kernel's code and data, and allocate the system page table.
 *	Called with mapping done by BATs. Page_size must already be set.
 *
 *	Parameters:
 *	first_avail	PA of address where we can allocate structures.
 */
void
pmap_bootstrap(unsigned int mem_size, vm_offset_t *first_avail)
{
	register struct mapping *mp;
	vm_offset_t addr;
	vm_size_t size;
	int i, num;
	unsigned int mask;
	vm_offset_t	first_used_addr;

	*first_avail = round_page(*first_avail);

	assert(PAGE_SIZE == PPC_PGBYTES);

	ppc_protection_init();

	/*
	 * Initialize kernel pmap
	 */
	kernel_pmap = &kernel_pmap_store;
#if	NCPUS > 1
	lock_init(&pmap_lock, FALSE, ETAP_VM_PMAP_SYS, ETAP_VM_PMAP_SYS_I);
#endif	/* NCPUS > 1 */
	simple_lock_init(&kernel_pmap->lock, ETAP_VM_PMAP_KERNEL);
	simple_lock_init(&free_pmap_lock, ETAP_VM_PMAP_FREE);

	kernel_pmap->ref_count = 1;
	kernel_pmap->space = PPC_SID_KERNEL;
	kernel_pmap->next = PMAP_NULL;

	/*
	 * Allocate: (from first_avail up)
	 *      aligned to its own size:
         *        hash table (for mem size 2**x, allocate 2**(x-10) entries)
	 *      optional on devices being used, aligned to 256Kbyte boundary
	 *        dma buffer, of size 256K
	 * (if hash_table_size is > DMA_BUFFER_ALIGNMENT then we reserve
	 *  dma buffer first in order to keep alignment)
	 *	physical page entries (1 per physical page)
	 *	physical -> virtual mappings (for multi phys->virt mappings)
	 */
	/* hash_table_size must be a power of 2, recommended sizes are
	 * taken from PPC601 User Manual, table 6-19. We take the next
	 * highest size if the memsize is not a power of two.
	 * TODO NMGS make this configurable at boot time.
	 */
	num = sizeof(pte_t) * (mem_size >> 10);

	for (hash_table_size = 64 * 1024;	/* minimum size = 64Kbytes */
	     hash_table_size < num; 
	     hash_table_size *= 2)
		continue;

	/* DMA BUFFER MUST BE aligned to its size
	 * HASH TABLE MUST BE aligned to its size */

	if (hash_table_size > powermac_info.dma_buffer_alignment)
		addr = (*first_avail +
			(hash_table_size-1)) & ~(hash_table_size-1);
	else
		addr = (*first_avail +
			(powermac_info.dma_buffer_alignment-1)) &
				~(powermac_info.dma_buffer_alignment-1);

	/* TODO NMGS - we step over the end of a read/write page here,
	 * TODO NMGS - in theory we could use this for mappings.
	 */
	if (round_page(*first_avail) + PPC_PGBYTES < round_page(addr)) {
		/* We are stepping over at least one page here, so
		 * add this region to the free regions so that it can
		 * be allocated by pmap_steal
		 */
		free_regions[free_regions_count].start = 
			round_page(*first_avail) + PPC_PGBYTES;
		free_regions[free_regions_count].end = round_page(addr);

		avail_remaining += (free_regions[free_regions_count].end -
				    free_regions[free_regions_count].start) /
					    PPC_PGBYTES;
#if DEBUG
		DPRINTF(("ADDED FREE REGION from 0x%08x to 0x%08x, avail_remaining = %d\n",free_regions[free_regions_count].start,free_regions[free_regions_count].end, avail_remaining));
#endif /* DEBUG */
		free_regions_count++;
	}

	num = atop(mem_size);	/* num now holds mem_size in pages */

	/* size of all structures that we're going to allocate */

	size = (vm_size_t) (hash_table_size + powermac_info.dma_buffer_size +
			    sizeof(struct phys_entry) * num +
			    (sizeof(struct mapping) *
			     hash_table_size / sizeof(pte_t)));

	size = round_page(size);

	/* Zero everything - this also invalidates the hash table entries */
	bzero((char *)addr, size);

	/* Set up some pointers to our new structures */

	/* from here,  addr points to the next free address */
	
	first_used_addr = addr;	/* remember where we started */

	/* Set up hash table address and dma buffer address, keeping
	 * alignment. These mappings are all 1-1,  so dma_r == dma_v
	 * 
	 * If hash_table_size == dma_buffer_alignment, then put hash_table
	 * first, since dma_buffer_size may be smaller than alignment, but
	 * hash table alignment==hash_table_size.
	 */
	if (hash_table_size >= powermac_info.dma_buffer_alignment) {
		hash_table_base = addr;
		addr += hash_table_size;

		powermac_info.dma_buffer_virt = addr;
		powermac_info.dma_buffer_phys = addr;

		addr += powermac_info.dma_buffer_size;
	} else {
		powermac_info.dma_buffer_virt = addr;
		powermac_info.dma_buffer_phys = addr;

		addr += powermac_info.dma_buffer_size;

		hash_table_base = addr;
		addr += hash_table_size;
	}
	assert((hash_table_base & (hash_table_size-1)) == 0);
	assert((powermac_info.dma_buffer_size == 0) ||
		(powermac_info.dma_buffer_virt &
		(powermac_info.dma_buffer_alignment-1)) == 0);

	/* phys_table is static to help debugging,
	 * this variable is no longer actually used
	 * outside of this scope
	 */

	phys_table = (struct phys_entry *) addr;

	for (i = 0; i < mem_regions_count; i++) {
		mem_regions[i].phys_table = phys_table;
		phys_table = phys_table +
			atop(mem_regions[i].end - mem_regions[i].start);
	}

	/* restore phys_table for debug */
	phys_table = (struct phys_entry *) addr;

	addr += sizeof(struct phys_entry) * num;

	/* Initialise the registers necessary for supporting the hashtable */
	hash_table_init(hash_table_base, hash_table_size);

	/* Initialise the physical table mappings */ 
	for (i = 0; i < num; i++) {
		queue_init(&phys_table[i].phys_link);
		phys_table[i].pte1.bits.wimg = PTE_WIMG_DEFAULT;
	}

	/*
	 * Remaining space is for mapping entries.  Chain them
	 * together (XXX can't use a zone
	 * since zone package hasn't been initialized yet).
	 */

	num = (size - (addr - first_used_addr)) /
		sizeof(struct mapping);

	mp = (struct mapping *) addr;

#if DEBUG
	DPRINTF(("mapping base, top, entries = 0x%08x, 0x%08x, %d\n",
		 mp, mp+num, num));
#endif /* DEBUG */

	queue_init(&free_mapping);
	simple_lock_init(&free_mapping_lock, ETAP_VM_PMAP_FREE);

	firstmap = mp;
	while (num--) {
		queue_enter(&free_mapping, mp, struct mapping *, phys_link);
		mp++;
	}
	lastmap = mp-1;

	/*
         * Map the memory we grabbed correctly using the page tables so
	 * that the BAT3 register becomes redundant.
	 */

#if DEBUG
DPRINTF(("mapping memory tables from 0x%08x to 0x%08x, to address 0x%08x\n",
	 first_used_addr, round_page(first_used_addr+size),
	 first_used_addr));
#endif /* DEBUG */

	pmap_map(first_used_addr,
		 first_used_addr,
		 round_page(first_used_addr+size),
		 VM_PROT_READ | VM_PROT_WRITE);

	/* Remap the DMA buffer region as a board (ie. non-cached) */
	pmap_map_bd(powermac_info.dma_buffer_virt,
		    powermac_info.dma_buffer_phys,
		    powermac_info.dma_buffer_virt +
		    	powermac_info.dma_buffer_size,
		    VM_PROT_READ | VM_PROT_WRITE);

	*first_avail = round_page(lastmap);

	/* All the rest of memory is free - add it to the free
	 * regions so that it can be allocated by pmap_steal
	 */
	free_regions[free_regions_count].start = *first_avail;
	free_regions[free_regions_count].end = mem_regions[0].end;

	avail_remaining += (free_regions[free_regions_count].end -
			    free_regions[free_regions_count].start) /
				    PPC_PGBYTES;

#if DEBUG
	DPRINTF(("ADDED FREE REGION from 0x%08x to 0x%08x, avail_remaining = %d\n",free_regions[free_regions_count].start,free_regions[free_regions_count].end, avail_remaining));
#endif /* DEBUG */

	free_regions_count++;

	current_free_region = 0;

	avail_next = free_regions[current_free_region].start;
}

/*
 * pmap_init(spa, epa)
 *	finishes the initialization of the pmap module.
 *	This procedure is called from vm_mem_init() in vm/vm_init.c
 *	to initialize any remaining data structures that the pmap module
 *	needs to map virtual memory (VM is already ON).
 */
/*ARGSUSED*/
void
pmap_init(void)
{
	vm_size_t s;

	s = sizeof(struct pmap);
	pmap_zone = zinit(s, 400*s, 4096, "pmap");	/* XXX */

	pmap_initialized = TRUE;
}

unsigned int pmap_free_pages(void)
{
	return avail_remaining;
}

boolean_t pmap_next_page(vm_offset_t *addrp)
{
	/* Non optimal, but only used for virtual memory startup.
         * Allocate memory from a table of free physical addresses
	 * If there are no more free entries, too bad. We have two
	 * tables to look through, free_regions[] which holds free
	 * regions from inside mem_regions[0], and the others...
	 * mem_regions[1..]
         */
	/* current_free_region indicates the next free entry,
	 * if it's less than free_regions_count, then we're still
	 * in free_regions, otherwise we're in mem_regions
	 */

	if (current_free_region >= free_regions_count) {
		/* We're into the mem_regions, handle this
		 * separately to free_regions
		 */

		int current_mem_region = current_free_region -
					 free_regions_count + 1;
		if (current_mem_region > mem_regions_count)
			return FALSE;
		*addrp = avail_next;
		avail_next += PAGE_SIZE;
		avail_remaining--;
		if (avail_next >= mem_regions[current_mem_region].end) {
			current_free_region++;
			current_mem_region++;
			avail_next = mem_regions[current_mem_region].start;
#if DEBUG
			DPRINTF(("pmap_next_page : next region start=0x%08x\n",avail_next));
#endif /* DEBUG */
		}
		return TRUE;
	}
	
	/* We're in the free_regions, allocate next page and increment
	 * counters
	 */
	*addrp = avail_next;

	avail_next += PAGE_SIZE;
	avail_remaining--;

	if (avail_next >= free_regions[current_free_region].end) {
		current_free_region++;
		if (current_free_region < free_regions_count)
			avail_next = free_regions[current_free_region].start;
		else
			avail_next = mem_regions[current_free_region -
						 free_regions_count + 1].start;
#if DEBUG
		printf("pmap_next_page : next region start=0x%08x\n",avail_next);
#endif 
	}
	return TRUE;
}

void pmap_virtual_space(
	vm_offset_t *startp,
	vm_offset_t *endp)
{
	/* give first and last virtual addresses available to kernel allocs */
	/* Give first address above which free space is guaranteed, ignoring
         * the holes below. In this case, the memory is mapped :
	 *
	 *			PHYSICAL	VIRTUAL
	 *   KERNEL TEXT 	0-1M		8-9M
	 *   VIDEO		1-2M		**** (9-10)
	 *   KERNEL DATA	2-xxx		2-xxx
	 *   IO stuff           > 0x50000000    1-1
	 *
	 * so the first guaranteed free virtual address is above the video
         */
	*startp = VPDM_BASEADDR + VPDM_FRAMEBUF_MAX_SIZE;
	*endp   = VM_MAX_KERNEL_ADDRESS;
}

/*
 * pmap_create
 *
 * Create and return a physical map.
 *
 * If the size specified for the map is zero, the map is an actual physical
 * map, and may be referenced by the hardware.
 *
 * If the size specified is non-zero, the map will be used in software 
 * only, and is bounded by that size.
 */
pmap_t
pmap_create(vm_size_t size)
{
	pmap_t pmap;
	int s;
	space_t sid;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_create(size=%x)%c", size, size ? '\n' : ' ');
#endif

	/*
	 * A software use-only map doesn't even need a pmap structure.
	 */
	if (size)
		return(PMAP_NULL);

	/* 
	 * If there is a pmap in the pmap free list, reuse it. 
	 */
	s = splvm();
	simple_lock(&free_pmap_lock);
	if (free_pmap != PMAP_NULL) {
		pmap = free_pmap;
		free_pmap = pmap->next;
	} else
		pmap = PMAP_NULL;
	simple_unlock(&free_pmap_lock);
	splx(s);

	/*
	 * Couldn't find a pmap on the free list, try to allocate a new one.
	 */
	if (pmap == PMAP_NULL) {
		pmap = (pmap_t) zalloc(pmap_zone);
		if (pmap == PMAP_NULL)
			return(PMAP_NULL);

		/*
		 * Allocate space IDs for the pmap.
		 * If all are allocated, there is nothing we can do.
		 */
		s = splvm();
		/* If sid_counter == MAX_SID, we've allocated
		 * all the possible SIDs and they're all live,
		 * we can't carry on, but this is HUGE, so don't
		 * expect it in normal operation.
		 */
		assert(sid_counter != SID_MAX);

		/* Try to spread out the sid's through the possible
		 * name space to improve the hashing heuristics
		 */
		sid_counter = (sid_counter + PPC_SID_PRIME) & PPC_SID_MASK;
		splx(s);

		/* 
		 * Initialize the sids
		 */
		pmap->space = sid_counter;
		simple_lock_init(&pmap->lock, ETAP_VM_PMAP);
	}
	pmap->ref_count = 1;
	pmap->next = PMAP_NULL;
	pmap->stats.resident_count = 0;
	pmap->stats.wired_count = 0;
#if DEBUG
	if (pmdebug & PDB_USER)
		printf("-> %x, space id = %d\n", pmap, pmap->space);
#endif
	return(pmap);
}

/* 
 * pmap_destroy
 * 
 * Gives up a reference to the specified pmap.  When the reference count 
 * reaches zero the pmap structure is added to the pmap free list.
 *
 * Should only be called if the map contains no valid mappings.
 */
void
pmap_destroy(pmap_t pmap)
{
	int ref_count;
	spl_t s;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_destroy(pmap=%x)\n", pmap);
#endif

	if (pmap == PMAP_NULL)
		return;

	s = splvm();
	simple_lock(&pmap->lock);
	ref_count = --pmap->ref_count;
	simple_unlock(&pmap->lock);
	splx(s);

	if (ref_count < 0)
		panic("pmap_destroy(): ref_count < 0");
	if (ref_count > 0)
		return;
	/* 
	 * Add the pmap to the pmap free list. 
	 */
	s = splvm();
	simple_lock(&free_pmap_lock);
	pmap->next = free_pmap;
	free_pmap = pmap;
	simple_unlock(&free_pmap_lock);
	splx(s);
}

/*
 * pmap_reference(pmap)
 *	gains a reference to the specified pmap.
 */
void
pmap_reference(pmap_t pmap)
{
	spl_t s;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_reference(pmap=%x)\n", pmap);
#endif

	if (pmap != PMAP_NULL) {
		s = splvm();
		simple_lock(&pmap->lock);
		pmap->ref_count++;
		simple_unlock(&pmap->lock);
		splx(s);
	}
}

/*
 * pmap_remove(pmap, s, e)
 *	unmaps all virtual addresses v in the virtual address
 *	range determined by [s, e) and pmap.
 *	s and e must be on machine independent page boundaries and
 *	s must be less than or equal to e.
 */
void
pmap_remove(
	    pmap_t pmap,
	    vm_offset_t sva,
	    vm_offset_t eva)
{
	space_t space;
	struct mapping *mp;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_remove(pmap=%x, sva=%x, eva=%x)\n",
		       pmap, sva, eva);
#endif

	if (pmap == PMAP_NULL)
		return;

	space = pmap->space;

	/* It is just possible that eva might have wrapped around to zero,
	 * and sometimes we get asked to liberate something of size zero
	 * even though it's dumb (eg. after zero length read_overwrites)
	 */
	assert(eva >= sva);

	/* We liberate addresses from high to low, since the stack grows
	 * down. This means that we won't need to test addresses below
	 * the limit of stack growth
	 */
	while (pmap->stats.resident_count && (eva > sva)) {
		eva -= PAGE_SIZE;
		if (mp = pmap_find_mapping(space, eva)) {
			pmap_free_mapping(mp);
			pmap->stats.resident_count--;
		}
	}
}

/*
 *	Routine:
 *		pmap_page_protect
 *
 *	Function:
 *		Lower the permission for all mappings to a given page.
 */
void
pmap_page_protect(
	vm_offset_t pa,
	vm_prot_t prot)
{
	register struct phys_entry *pp;
	register struct mapping *mp;
	unsigned pteprot;
	boolean_t remove;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_page_protect(pa=%x, prot=%x)\n", pa, prot);
#endif

	switch (prot) {
	case VM_PROT_READ:
	case VM_PROT_READ|VM_PROT_EXECUTE:
		remove = FALSE;
		break;
	case VM_PROT_ALL:
		return;
	default:
		remove = TRUE;
		break;
	}

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	if (remove) {
		while (!queue_empty(&pp->phys_link)) {
			mp = (struct mapping *) queue_first(&pp->phys_link);
			mp->pmap->stats.resident_count--;
			pmap_free_mapping(mp);
		}
		return;
	}

	/*
	 * Modify mappings if necessary.
	 */
	queue_iterate(&pp->phys_link, mp, struct mapping *, phys_link) {
		/*
		 * Compare new protection with old to see if
		 * anything needs to be changed.
		 */
		pteprot = prot_bits[prot];
		if ((mp->pte->pte1.bits.protection) != pteprot) {

			/*
			 * Purge the current TLB entry for that
			 * virtual address (if any) to force saving
			 * of change/reference bits, followed by a
			 * fault and reload with the new protection.
			 */
			invalidate_tlb(mp->space,
				       mp->vm_info.bits.page << PPC_PGSHIFT);

			INVALIDATE_CACHED_PTE(mp->pte);

			mp->pte->pte1.bits.protection = pteprot;

			WRITEBACK_CACHED_PTE(mp->pte);
		}
		pp->pte1.word |= mp->pte->pte1.word;
	}
	if ((PROCESSOR_VERSION != PROCESSOR_VERSION_601) &&
	    (pp->pte1.bits.changed)) {
		flush_cache((vm_offset_t)(pa & (PAGE_SIZE-1)),
			    PAGE_SIZE);
	}
}

boolean_t
pmap_verify_free(vm_offset_t pa)
{
	struct phys_entry *pp;
	boolean_t result;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_verify_free(pa=%x)\n", pa);
#endif

	if (!pmap_initialized)
		return(TRUE);

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return(FALSE);

	return(queue_empty(&pp->phys_link));
}

/*
 * pmap_protect(pmap, s, e, prot)
 *	changes the protection on all virtual addresses v in the 
 *	virtual address range determined by [s, e] and pmap to prot.
 *	s and e must be on machine independent page boundaries and
 *	s must be less than or equal to e.
 */
void
pmap_protect(
	     pmap_t pmap,
	     vm_offset_t sva, 
	     vm_offset_t eva,
	     vm_prot_t prot)
{
	register struct mapping *mp;
	register struct phys_entry *pp;
	unsigned pteprot;
	space_t space;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_protect(pmap=%x, sva=%x, eva=%x, prot=%x)\n",
		       pmap, sva, eva, prot);
#endif

	if (pmap == PMAP_NULL)
		return;

	assert(sva <= eva);

	if (prot == VM_PROT_NONE) {
		pmap_remove(pmap, sva, eva);
		return;
	}
	if (prot & VM_PROT_WRITE)
		return;

	space = pmap->space;
	for ( ; sva < eva; sva += PAGE_SIZE) {
		mp = pmap_find_mapping(space, sva);
		if (mp == MAPPING_NULL)
			continue;
#if DEBUG
		if (pmap != mp->pmap)
			panic("protect: pmap mismatch");
#endif
		if (PROCESSOR_VERSION != PROCESSOR_VERSION_601)
			flush_cache((vm_offset_t)(mp->pte->pte1.word &
						  (PAGE_SIZE-1)),
				    PAGE_SIZE);
		/*
		 * Determine if mapping is changing.
		 * If not, nothing to do.
		 */
		pteprot = prot_bits[prot];
		if (mp->pte->pte1.bits.protection == pteprot)
			continue;
		
		/*
		 * Purge the current TLB entry (if any) to force
		 * any modifications to changed/referenced bits, and so
		 * that future references will fault and reload with the
		 * new protection.
		 */
		invalidate_tlb(space, sva);

		/* Change just the protection */
		INVALIDATE_CACHED_PTE(mp->pte);
		mp->pte->pte1.bits.protection = pteprot;
		WRITEBACK_CACHED_PTE(mp->pte);
	}
}

/*
 * pmap_enter
 *
 * Create a translation for the virtual address (virt) to the physical
 * address (phys) in the pmap with the protection requested. If the
 * translation is wired then we can not allow a page fault to occur
 * for this mapping.
 *
 * NB: This is the only routine which MAY NOT lazy-evaluate
 *     or lose information.  That is, this routine must actually
 *     insert this page into the given map NOW.
 */
void
pmap_enter(pmap_t pmap, vm_offset_t va, vm_offset_t pa, vm_prot_t prot,
	   boolean_t wired)
{
	register struct mapping *mp;
	struct phys_entry *pp,*old_pp;
	pte_t *pte;
	space_t space;
	boolean_t was_wired;
	int x;

#if DEBUG
	if ((wired && (pmdebug & (PDB_WIRED))) ||
	     (pmdebug & (PDB_USER|PDB_ENTER)))
		printf("pmap_enter(pmap=%x, va=%x, pa=%x, prot=%x, wire=%x)\n",
		       pmap, va, pa, prot, wired);
#endif
        if (pmap == PMAP_NULL)
                return;

	/* TODO NMGS - take lock on physentry? */
	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	/* Find the pte associated with the virtual address/space ID */
	space = pmap->space;

	/* TODO NMGS - take lock on pte?! */
	pte = find_or_allocate_pte(space,va,TRUE);

	assert(pte != PTE_NULL);
	/* Even if the pte is new, pte0 should be set up for the new mapping,
	 * but with the valid bit set to FALSE
	 */
	assert((pte->pte0.word & 0x7fffffff) != PTE_EMPTY);

	if (pte->pte0.bits.valid == TRUE) {

		/* mapping is already valid, thus virtual address was
		 * already mapped somewhere...
		 */

		if (((physical_addr_t *) &pa)->bits.page_no !=
	    	    pte->pte1.bits.phys_page) {

			/* The current mapping is to a different
			 * physical addr - release the mapping
			 * before mapping the new address. 
			 */

#if DEBUG
			if (pmdebug & (PDB_USER|PDB_ENTER))
				printf("Virt 0x%08x was mapped to different address 0x%08x, changing mapping\n",va,((physical_addr_t *) &pa)->bits.page_no * PPC_PGBYTES);
#endif

			/* Take advantage of the fact that pte1 can be
			 * considered as a pointer to an address in
			 * the physical page when looking for physentry.
			 */

			old_pp = pmap_find_physentry(
					     (vm_offset_t)(pte->pte1.word));

			queue_iterate(&old_pp->phys_link, mp,
				      struct mapping *, phys_link) {
				if (mp->pte == pte)
					break;
			}

			/* pte entry must have good reverse mapping */
			assert(mp != NULL);
			assert(mp->pte == pte);
			assert(((mp->vm_info.bits.page >> 10) & 0x3f) ==
			       mp->pte->pte0.bits.page_index);

			/* Wired/not wired is taken care of later on
			 * in this function. We can remove wired mappings.
			 */

			/* Do the equivalent of pmap_free_mapping followed
			 * by a pmap_enter_mapping, recycling the pte and
			 * the mapping
			 */

			invalidate_tlb(space, va);

			/* keep track of changed and referenced at old
			 * address
			 */
			
			INVALIDATE_CACHED_PTE(mp->pte);
			old_pp->pte1.word |= mp->pte->pte1.word;

			x = splvm();
			queue_remove(&old_pp->phys_link, mp,
				     struct mapping *, phys_link);

			mp->space = space;
			mp->vm_info.bits.page  = (va >> PPC_PGSHIFT);

			assert(mp->pte == pte);

			pte->pte0.bits.valid = FALSE;

			queue_enter_first(&pp->phys_link, mp,
					  struct mapping *, phys_link);

			splx(x);

			assert(mp->pmap == pmap);

			/* Set up pte as it would have been done by
			 * find_or_allocate_pte() + pmap_enter_mapping()
			 */
			pte->pte0.bits.segment_id = 
				(space << 4) | (va >> 28);
			pte->pte0.bits.page_index =
				((va_abbrev_t *)&va)->page_index;
			pte->pte1.word = 0;
			pte->pte1.bits.phys_page = pa >> PPC_PGSHIFT;
			pte->pte1.bits.protection = prot_bits[prot];

			pte->pte1.bits.wimg = pp->pte1.bits.wimg; /* default */

			/* make sure those pte1 writes have completed */
			isync();
			pte->pte0.bits.valid = TRUE;

		} else {
			/* The current mapping is the same as the one
			 * we're being asked to make, find the mapping
			 * structure that goes along with it.
			 */
			queue_iterate(&pp->phys_link, mp,
				      struct mapping *, phys_link) {
				if (mp->pte == pte)
					break;
			}
			/* mapping struct must exist */
			assert(mp != NULL);
			assert(mp->pte == pte);

			/* invalidate tlb for previous mapping to make sure
			 * that we don't trample on changed/referenced bits
			 */
			invalidate_tlb(space, va);

			INVALIDATE_CACHED_PTE(pte); /* Force re-read */

#if DEBUG
			if ((wired && (pmdebug & (PDB_WIRED))) ||
			    (pmdebug & (PDB_USER|PDB_ENTER))) {
				printf("address already mapped, changing prots or wiring\n");
				if (pte->pte1.bits.protection !=
				    prot_bits[prot])
					printf("Changing PROTS\n");
				if (wired != mp->vm_info.bits.wired)
				       printf("changing WIRING\n");
			}
#endif
		}
	} else {
		/* There was no pte match, a new entry was created
                 * (marked as invalid). We make sure that a new mapping
		 * structure is allocated as it's partner.
		 */
		mp = MAPPING_NULL;
		assert(pte->pte0.word != PTE_EMPTY);
	}

	/* Once here, pte contains a valid pointer to a structure
	 * that we may use to map the virtual address (which may
	 * already map this (and be valid) or something else (invalid)
	 * and mp contains either MAPPING_NULL (new mapping) or
	 * a pointer to the old mapping which may need it's
	 * privileges modified.
	 */

	/* set up the protection bits -
	 * MUSTN'T INVALIDATE PTE here, might contain unflushed mapping
	 */

	pte->pte1.bits.protection = prot_bits[prot];
	WRITEBACK_CACHED_PTE(pte);

	if (mp == MAPPING_NULL) {
		/*
		 * Mapping for this virtual address doesn't exist.
		 * Get a mapping structure and, and fill it in,
		 * updating the pte that we already have, and
		 * making it valid
		 */
		mp = pmap_enter_mapping(pmap, space, va, pa, pte, prot, pp);
		pmap->stats.resident_count++;
		was_wired = FALSE;/* indicate that page was never wired */
	} else {
		
		/*
		 * We are just changing the protection of a current mapping.
		 */
		was_wired = mp->vm_info.bits.wired;
	}

	/* if wired status has changed, update stats and change bit */
	if (was_wired != wired) {
		pmap->stats.wired_count += wired ? 1 : -1;
		mp->vm_info.bits.wired = wired;
#if DEBUG
		if (pmdebug & (PDB_WIRED))
			printf("changing wired status to : %s\n",
			       wired ? "TRUE" : "FALSE");
#endif /* DEBUG */
	}

	/* Synchronise caches - NEEDED, even if non-execute (512k L2 probs) */
	if (PROCESSOR_VERSION != PROCESSOR_VERSION_601)
		flush_cache(pa, PAGE_SIZE);


	/* Belt and braces check to make sure we didn't give a bogus map,
	 * this map is given once when mapping the kernel
	 */
	assert((pte->pte1.bits.phys_page != 0) ||
	       (kgdb_kernel_in_pmap == FALSE));
	
#if DEBUG
	if (pmdebug & PDB_MASSIVE) {
		register struct mapping *mp2;
		queue_iterate(&free_mapping, mp2,
			      struct mapping *, phys_link) {
			assert(mp2->phys_link.next != NULL);
		}
	}
#endif /* DEBUG */	
#if DEBUG
	if (pmdebug & (PDB_USER|PDB_ENTER))
		printf("leaving pmap_enter\n");
#endif
}

/*
 *	Routine:	pmap_change_wiring
 *	Function:	Change the wiring attribute for a map/virtual-address
 *			pair.
 *	In/out conditions:
 *			The mapping must already exist in the pmap.
 *
 * Change the wiring for a given virtual page. This routine currently is 
 * only used to unwire pages and hence the mapping entry will exist.
 */
void
pmap_change_wiring(
	register pmap_t	pmap,
	vm_offset_t	va,
	boolean_t	wired)
{
	register struct mapping *mp;
	boolean_t waswired;

#if DEBUG
	if (pmdebug & (PDB_USER | PDB_WIRED))
		printf("pmap_change_wiring(pmap=%x, va=%x, wire=%x)\n",
		       pmap, va, wired);
#endif
	if (pmap == PMAP_NULL)
		return;

	if ((mp = pmap_find_mapping(pmap->space, va)) == MAPPING_NULL)
		panic("pmap_change_wiring: can't find mapping entry");

	waswired = mp->vm_info.bits.wired;
	if (wired && !waswired) {
		mp->vm_info.bits.wired = TRUE;
		pmap->stats.wired_count++;
	} else if (!wired && waswired) {
		mp->vm_info.bits.wired = FALSE;
		pmap->stats.wired_count--;
	}
}

/*
 * pmap_extract(pmap, va)
 *	returns the physical address corrsponding to the 
 *	virtual address specified by pmap and va if the
 *	virtual address is mapped and 0 if it is not.
 */
vm_offset_t
pmap_extract(
	pmap_t pmap,
	vm_offset_t va)
{
	pte_t *pte;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_extract(pmap=%x, va=%x)\n", pmap, va);
#endif

	pte = find_or_allocate_pte(pmap->space, trunc_page(va), FALSE);
	if (pte == NULL)
		return(0);
	return trunc_page(pte->pte1.word) | (va & page_mask);
}

/*
 *	pmap_attributes:
 *
 *	Set/Get special memory attributes
 *
 */
kern_return_t
pmap_attribute(pmap, address, size, attribute, value)
	pmap_t		pmap;
	vm_offset_t	address;
	vm_size_t	size;
	vm_machine_attribute_t	attribute;
	vm_machine_attribute_val_t* value;	
{
	vm_offset_t 	sva, eva;
	vm_offset_t	addr;
	pte_t		*pte;
	kern_return_t	ret;

	if (attribute != MATTR_CACHE)
		return KERN_INVALID_ARGUMENT;

	/* We can't get the caching attribute for more than one page
	 * at a time
	 */
	if ((*value == MATTR_VAL_GET) &&
	    (trunc_page(address) != trunc_page(address+size-1)))
		return KERN_INVALID_ARGUMENT;

	if (pmap == PMAP_NULL)
		return KERN_SUCCESS;

	sva = trunc_page(address);
	eva = round_page(address + size);
	ret = KERN_SUCCESS;

	simple_lock(&pmap->lock);

	switch (*value) {
	case MATTR_VAL_CACHE_FLUSH:	/* flush from all caches */
	case MATTR_VAL_DCACHE_FLUSH:	/* flush from data cache(s) */
	case MATTR_VAL_ICACHE_FLUSH:	/* flush from instr cache(s) */
		sva = trunc_page(sva);
		for (; sva < eva; sva += PAGE_SIZE) {
			if ((addr = pmap_extract(pmap, sva)) == 0)
				continue;
			flush_cache(addr, PAGE_SIZE);
		}
		break;

	case MATTR_VAL_GET:		/* return current value */
	case MATTR_VAL_OFF:		/* turn attribute off */
	case MATTR_VAL_ON:		/* turn attribute on */
	default:
		ret = KERN_INVALID_ARGUMENT;
		break;
	}
	simple_unlock(&pmap->lock);

	return ret;
}

/*
 * pmap_collect
 * 
 * Garbage collects the physical map system for pages that are no longer used.
 */
/* ARGSUSED */
void
pmap_collect(pmap_t pmap)
{
#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_collect(pmap=%x)\n", pmap);
#endif
}

/*
 *	Routine:	pmap_activate
 *	Function:
 *		Binds the given physical map to the given
 *		processor, and returns a hardware map description.
 */
/* ARGSUSED */
void
pmap_activate(
	pmap_t pmap,
	thread_t th,
	int which_cpu)
{
#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_activate(pmap=%x, th=%x, cpu=%x)\n",
		       pmap, th, which_cpu);
#endif
}

/* ARGSUSED */
void
pmap_deactivate(
	pmap_t pmap,
	thread_t th,
	int which_cpu)
{
#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_activate(pmap=%x, th=%x, cpu=%x)\n",
		       pmap, th, which_cpu);
#endif
}

#if DEBUG

/*
 * pmap_zero_page
 * pmap_copy page
 * 
 * are implemented in movc.s, these
 * are just wrappers to help debugging
 */

extern void pmap_zero_page_assembler(vm_offset_t p);
extern void pmap_copy_page_assembler(vm_offset_t src, vm_offset_t dst);

/*
 * pmap_zero_page(pa)
 *
 * pmap_zero_page zeros the specified (machine independent) page pa.
 */
void
pmap_zero_page(
	vm_offset_t p)
{
	register struct mapping *mp;
	register struct phys_entry *pp;

	if (pmdebug & (PDB_USER|PDB_ZERO))
		printf("pmap_zero_page(pa=%x)\n", p);

	/*
	 * XXX can these happen?
	 */
	if (pmap_find_physentry(p) == PHYS_NULL)
		panic("zero_page: physaddr out of range");

	/* TODO NMGS optimisation - if page has no mappings, set bit in
	 * physentry to indicate 'needs zeroing', then zero page in
	 * pmap_enter or pmap_copy_page if bit is set. This keeps the
	 * cache 'hot'.
	 */

	pmap_zero_page_assembler(p);
}

/*
 * pmap_copy_page(src, dst)
 *
 * pmap_copy_page copies the specified (machine independent)
 * page from physical address src to physical address dst.
 *
 * We need to invalidate the cache for address dst before
 * we do the copy. Apparently there won't be any mappings
 * to the dst address normally.
 */
void
pmap_copy_page(
	vm_offset_t src,
	vm_offset_t dst)
{
	register struct phys_entry *pp;
	register struct mapping *mp;

	if (pmdebug & (PDB_USER|PDB_COPY))
		printf("pmap_copy_page(spa=%x, dpa=%x)\n", src, dst);
	if (pmdebug & PDB_COPY)
		printf("pmap_copy_page: phys_copy(%x, %x, %x)\n",
		       src, dst, PAGE_SIZE);

	pmap_copy_page_assembler(src, dst);
}
#endif /* DEBUG */

/*
 * pmap_pageable(pmap, s, e, pageable)
 *	Make the specified pages (by pmap, offset)
 *	pageable (or not) as requested.
 *
 *	A page which is not pageable may not take
 *	a fault; therefore, its page table entry
 *	must remain valid for the duration.
 *
 *	This routine is merely advisory; pmap_enter()
 *	will specify that these pages are to be wired
 *	down (or not) as appropriate.
 *
 *	(called from vm/vm_fault.c).
 */
/* ARGSUSED */
void
pmap_pageable(
	pmap_t		pmap,
	vm_offset_t	start,
	vm_offset_t	end,
	boolean_t	pageable)
{
#if DEBUG
	if (pmdebug & (PDB_USER | PDB_WIRED))
		printf("pmap_pageable(pmap=%x, sva=%x, eva=%x, pageable=%s)\n",
		       pmap, start, end, pageable ? "TRUE" : "FALSE");
#endif
}

/*
 * pmap_modify_pages(pmap, s, e)
 *	sets the modified bit on all virtual addresses v in the 
 *	virtual address range determined by [s, e] and pmap,
 *	s and e must be on machine independent page boundaries and
 *	s must be less than or equal to e.
 */
void
pmap_modify_pages(
	     pmap_t pmap,
	     vm_offset_t sva, 
	     vm_offset_t eva)
{
	register pte_t	*pte;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_modify_pages(pmap=%x, sva=%x, eva=%x)\n",
		       pmap, sva, eva);
#endif

	if (pmap == PMAP_NULL)
		return;

	for ( ; sva < eva; sva += PAGE_SIZE) {
		pte = find_or_allocate_pte(pmap->space, trunc_page(sva),FALSE);

		if (pte != PTE_NULL) {
			/*
			 * It's cheaper to modify the pte than the
			 * phys_entry, but make sure a tlb writeback
			 * won't cause the changed bit to be lost.
			 */
			invalidate_tlb(pmap->space, sva);

			INVALIDATE_CACHED_PTE(pte);
			pte->pte1.bits.changed = TRUE;
			WRITEBACK_CACHED_PTE(pte);

		} else {
			printf("WARNING : pte missing for pmap_modify_pages\n");
		}
	}
}

/*
 * pmap_clear_modify(phys)
 *	clears the hardware modified ("dirty") bit for one
 *	machine independant page starting at the given
 *	physical address.  phys must be aligned on a machine
 *	independant page boundary.
 */
void
pmap_clear_modify(vm_offset_t pa)
{
	register struct phys_entry *pp;
	register struct mapping *mp;
	register space_t space;
	register vm_offset_t offset;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_clear_modify(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	/*
	 * invalidate any pte entries and remove them from tlb
	 * TODO might it be faster to wipe all tlb entries outside loop?
	 */

	/* Remove any reference to modified to the physical page */
	pp->pte1.bits.changed = FALSE;

	/* mark PTE entries as having no modifies, and flush tlbs */
	queue_iterate(&pp->phys_link, mp,
		      struct mapping *, phys_link) {
		space = mp->space;
		offset = mp->vm_info.bits.page << PPC_PGSHIFT;

		invalidate_tlb(space,offset);

		/* Don't stomp on referenced */
		INVALIDATE_CACHED_PTE(mp->pte);
		mp->pte->pte1.bits.changed = FALSE;
		WRITEBACK_CACHED_PTE(mp->pte);
	}

	/* We must keep cache coherence
	 * between the data cache and the instruction cache, so
	 * we must flush the data and invalidate the instruction
	 * cache if they are not unified
	 */
	if (PROCESSOR_VERSION != PROCESSOR_VERSION_601)
		flush_cache(pa & (PAGE_SIZE-1), PAGE_SIZE);

}

/*
 * pmap_is_modified(phys)
 *	returns TRUE iff the given physical page has been modified 
 *	since the last call to pmap_clear_modify().
 */
boolean_t
pmap_is_modified(register vm_offset_t pa)
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_is_modified(pa=%x)\n", pa);
#endif
	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return(FALSE);

	/* Check to see if we've already noted this as modified */
	if (pp->pte1.bits.changed)
		return(TRUE);

	queue_iterate(&pp->phys_link, mp,
		      struct mapping *, phys_link) {
		/* Purge the current TLB entry for each mapping, to make
		 * sure that the mapping we have in memory reflects the
		 * true state of the entry
		 */
		invalidate_tlb(mp->space,
			       mp->vm_info.bits.page << PPC_PGSHIFT);

		INVALIDATE_CACHED_PTE(mp->pte);
		if (mp->pte->pte1.bits.changed) {
			pp->pte1.bits.changed = TRUE;
			return TRUE;
		}
	}
	
	return(FALSE);
}

/*
 * pmap_clear_reference(phys)
 *	clears the hardware referenced bit in the given machine
 *	independant physical page.  
 *
 */
void
pmap_clear_reference(vm_offset_t pa)
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_clear_reference(pa=%x)\n", pa);
#endif

	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return;

	/*
	 * Flush any pte entries to ensure no referenced bits are set
	 */
	queue_iterate(&pp->phys_link, mp,
		      struct mapping *, phys_link) {
		invalidate_tlb(mp->space,
			       mp->vm_info.bits.page << PPC_PGSHIFT);

		INVALIDATE_CACHED_PTE(mp->pte);
		mp->pte->pte1.bits.referenced = FALSE;
		WRITEBACK_CACHED_PTE(mp->pte);
	}

	/* mark physical entry as non-referenced too*/
	pp->pte1.bits.referenced = FALSE;
}

/*
 * pmap_is_referenced(phys)
 *	returns TRUE iff the given physical page has been referenced 
 *	since the last call to pmap_clear_reference().
 */
boolean_t
pmap_is_referenced(vm_offset_t pa)
{
	register struct phys_entry *pp;
	register struct mapping *mp;

#if DEBUG
	if (pmdebug & PDB_USER)
		printf("pmap_is_referenced(pa=%x)\n", pa);
#endif
	pp = pmap_find_physentry(pa);
	if (pp == PHYS_NULL)
		return(FALSE);

	if (pp->pte1.bits.referenced)
		return (TRUE);

	/* Check all mappings to see if any have made a reference */
	queue_iterate(&pp->phys_link, mp,
		      struct mapping *, phys_link) {
		/* Purge the current TLB entry for each mapping, to make
		 * sure that the mapping we have in memory reflects the
		 * true state of the entry
		 */
		invalidate_tlb(mp->space,
			       mp->vm_info.bits.page << PPC_PGSHIFT);

		INVALIDATE_CACHED_PTE(mp->pte);
		if (mp->pte->pte1.bits.referenced) {
			pp->pte1.bits.referenced = TRUE;
			return (TRUE);
		}
	}
	return(FALSE);
}

/*
 * Auxilliary routines for manipulating mappings
 */

/* static */
static struct mapping *
pmap_enter_mapping(pmap_t	pmap,
		   space_t	space,
		   vm_offset_t	va,
		   vm_offset_t	pa,
		   pte_t	*pte,
		   unsigned	prot,
		   struct phys_entry *pp)
{
	register struct mapping *mp;
	int x;

#if DEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_enter_mapping(pmap=%x, sp=%x, off=%x, &pte=%x, prot=%x)",
		       pmap, space, va, pte, prot);
#endif /* DEBUG */

	x = splvm();
	simple_lock(&free_mapping_lock);
	if (queue_empty(&free_mapping))
		pmap_reap_mappings();
	queue_remove_first(&free_mapping, mp, struct mapping *, phys_link);

	simple_unlock(&free_mapping_lock);
	splx(x);
	assert(mp != MAPPING_NULL);

	mp->space = space;
	mp->vm_info.bits.wired = FALSE;
	mp->vm_info.bits.page  = (va >> PPC_PGSHIFT);
	mp->pte  = pte;

	/* Set up pte -
	 *   pte0 already contains everything except the valid bit.
	 *   pte1 already contains the protection information.
	 */

	assert(pte->pte0.word != PTE_EMPTY);

	pte->pte1.bits.phys_page = pa >> PPC_PGSHIFT;

	pte->pte1.bits.wimg = pp->pte1.bits.wimg; /* Set default cachability*/

	/* make sure those pte1 writes above have completed */
	isync();
	pte->pte0.bits.valid = TRUE;		/* Validate pte entry */
	WRITEBACK_CACHED_PTE(pte);

	x = splvm();
	queue_enter_first(&pp->phys_link, mp, struct mapping *, phys_link);
	splx(x);

	mp->pmap = pmap;	/* TODO NMGS - why is this after entry? */

#if	DEBUG
	/* Assert that there are no other mappings with different
	 * cachability information, as this can freeze the machine
	 */
	{
		struct mapping *mp2;
		queue_iterate(&pp->phys_link, mp2,
			      struct mapping *, phys_link) {
			assert(mp2->pte->pte1.bits.wimg ==
			       pte->pte1.bits.wimg);
		}
	}
#endif	/* DEBUG */
#if DEBUG
	if (pmdebug & PDB_MAPPING)
		printf(" -> %x\n");
#endif
#if DEBUG
	if (pmdebug & PDB_MASSIVE) {
		register struct mapping *mp2;
		queue_iterate(&free_mapping, mp2,
			      struct mapping *, phys_link) {
			assert(mp2->phys_link.next != NULL);
		}
	}
#endif /* DEBUG */	
	return(mp);
}

/*
 * pmap_find_mapping(space, offset)
 *	Lookup the virtual address <space,offset> in the mapping "table".
 *	Returns a pointer to the mapping or NULL if none.
 *
 *	XXX what about MP locking?
 */
/* static */
struct mapping *
pmap_find_mapping(
	register space_t space,
	vm_offset_t offset)
{
	register struct mapping *mp;
	register struct phys_entry *pp;
	register pte_t *pte;

#if DEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_find_mapping(sp=%x, off=%x) -> ", space, offset);
#endif
	offset = trunc_page(offset);

	/* Locate the pte but don't allocate a new one if not found */
	pte = find_or_allocate_pte(space, offset, FALSE);

	if (pte == NULL) {
#if DEBUG
		if (pmdebug & PDB_MAPPING)
			printf("MAPPING_NULL\n");
#endif /* DEBUG */
		return MAPPING_NULL;
	}

	/* Take advantage of the fact that pte1 can be
	 * considered as a pointer to an address in
	 * the physical page when looking for physentry.
	 */

	pp = pmap_find_physentry((vm_offset_t)(pte->pte1.word));

	queue_iterate(&pp->phys_link, mp,
		      struct mapping *, phys_link) {
		if (mp->pte == pte)
			break;
	}
	assert(mp != MAPPING_NULL);	/* There must be mapping for a pte */
	assert(mp->pte == pte);
	assert(((mp->vm_info.bits.page >> 10) & 0x3f) ==
	       mp->pte->pte0.bits.page_index);

#if DEBUG
	if (pmdebug & PDB_MAPPING)
		printf("%x\n", mp);
#endif
	return(mp);
}

/* static */
void
pmap_free_mapping(register struct mapping *mp)
{
	register space_t space;
	register vm_offset_t offset;
	struct phys_entry *pp;
	int x;

#if DEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_free_mapping(mp=%x), pte at 0x%x (0x%08x,0x%08x)\n",
		       mp, mp->pte, mp->pte->pte0.word,mp->pte->pte1.word);
#endif

	space = mp->space;
	offset = mp->vm_info.bits.page << PPC_PGSHIFT;

	/*
	 * First flush cache and TLBs.
	 */
	pp = pmap_find_physentry((vm_offset_t)mp->pte->pte1.word);
	assert(pp != PHYS_NULL);

	invalidate_tlb(space, offset);

	/* Reflect back the modify and reference bits for the pager */
	INVALIDATE_CACHED_PTE(mp->pte);
	pp->pte1.word |= mp->pte->pte1.word;

	/*
	 * Now remove from the PTOV/VTOP lists and return to the free list.
	 * Note that we must block interrupts since they might cause TLB
	 * misses which would search (potentially inconsistant) lists.
	 */

	x = splvm();

#if 0
	/* one day we can get sophisticated with the pte handling */
	remove_pte(pte);
#else
	/* for now, just remove the mapping */
	mp->pte->pte0.word = PTE_EMPTY; /* Remove pte (v->p) lookup */
#if DEBUG
	mp->pte->pte1.word = 0;
#endif	
	WRITEBACK_CACHED_PTE(mp->pte);
#endif /* 0 */

	queue_remove(&pp->phys_link, mp, struct mapping *, phys_link);

	simple_lock(&free_mapping_lock);
	queue_enter(&free_mapping, mp, struct mapping *, phys_link);
	simple_unlock(&free_mapping_lock);
	splx(x);

#if DEBUG
	if (pmdebug & PDB_MASSIVE) {
		register struct mapping *mp2;
		queue_iterate(&free_mapping, mp2,
			      struct mapping *, phys_link) {
			assert(mp2->phys_link.next != NULL);
		}
	}
#endif /* DEBUG */	
}

/*
 * Deal with a hash-table pteg overflow. pmap_pteg_overflow is upcalled from
 * find_or_allocate_pte(), and assumes knowledge of the pteg hashing
 * structure
 */

static int pteg_overflow_counter=0;

static pte_t *pteg_pte_exchange(pte_t *pteg, pte0_t match);
pte_t *pmap_pteg_overflow(pte_t *primary_hash, pte0_t primary_match,
			  pte_t *secondary_hash, pte0_t secondary_match)
{
	pte_t *pte;
#if DEBUG
	int i;
	if (pmdebug & PDB_PTEG) {
		printf("pteg overflow for ptegs 0x%08x and 0x%08x\n",
		       primary_hash,secondary_hash);
		printf("PRIMARY\t\t\t\t\tSECONDARY\n");
		for (i=0; i<8; i++) {
			printf("0x%08x : (0x%08x,0x%08x)\t"
			       "0x%08x : (0x%08x,0x%08x)\n",
			       &primary_hash[i],
			       primary_hash[i].pte0.word,
			       primary_hash[i].pte1.word,
			       &secondary_hash[i],
			       secondary_hash[i].pte0.word,
			       secondary_hash[i].pte1.word);
			assert(primary_hash[i].pte0.word != PTE_EMPTY);
			assert(secondary_hash[i].pte0.word != PTE_EMPTY);
		}
	}
#endif
	/* First try to replace an entry in the primary PTEG */
	pte = pteg_pte_exchange(primary_hash, primary_match);
	if (pte != PTE_NULL)
		return pte;
	pte = pteg_pte_exchange(secondary_hash, secondary_match);
	if (pte != PTE_NULL)
		return pte;

	panic("both ptegs are completely wired down\n");
	return PTE_NULL;
}
			
/* Used by the pmap_pteg_overflow function to scan through a pteg */

static pte_t *pteg_pte_exchange(pte_t *pteg, pte0_t match)
{
	/* Try and replace an entry in a pteg */

	int i;
	register struct phys_entry *pp;
	register struct mapping *mp;
	pte_t *pte;

	/* TODO NMGS pteg overflow algorithm needs working on!
	 * currently we just loop around the PTEGs rejecting
	 * the first available entry.
	 */

	for (i=0; i < 8; i++) {

		pte = &pteg[pteg_overflow_counter];
		pteg_overflow_counter = (pteg_overflow_counter+1) % 8;

		/* all mappings must be valid otherwise we wouldn't be here */
		assert(pte->pte0.bits.valid == TRUE);

		/* Take advantage of the fact that pte1 can be
		 * considered as a pointer to an address in
		 * the physical page when looking for physentry.
		 */
		pp = pmap_find_physentry(pte->pte1.word);

		/* pte entry must have reverse mapping, otherwise we
		 * cannot remove it (it was entered by pmap_map_bd)
		 */
		if (pp == PHYS_NULL)
			continue;

		queue_iterate(&pp->phys_link, mp,
			      struct mapping *, phys_link) {
			if (mp->pte == pte)
				break;
		}
		assert(mp != NULL);
		assert(mp->pte == pte);

		/* if this entry isn't wired, isn't in the kernel space,
		 * and it doesn't have any special cachability attributes
		 * we can remove mapping without losing any information.
		 * XXX kernel space mustn't be touched as we may be doing
		 * XXX an I/O on a non-wired page, and we must not take
		 * XXX a page fault in an interrupt handler.
		 */
		/* TODO better implementation for cache info? */
		if (!mp->vm_info.bits.wired &&
		    mp->space != PPC_SID_KERNEL &&
		    mp->pte->pte1.bits.wimg == PTE_WIMG_DEFAULT) {
#if DEBUG
			if (pmdebug & PDB_PTEG) {
				printf("entry chosen at 0x%08x in %s PTEG\n",
				       pte,
				       pte->pte0.bits.hash_id ?
				       	"SECONDARY" : "PRIMARY");
				printf("throwing away mapping from sp 0x%x, "
				       "virtual 0x%08x to physical 0x%08x\n",
				       mp->space,
				       mp->vm_info.bits.page << PPC_PGSHIFT,
				       trunc_page(mp->pte->pte1.word));
				printf("pteg_overflow_counter=%d\n",
				       pteg_overflow_counter);
				printf("pte0.vsid=0x%08x\n",
				       pte->pte0.bits.segment_id);
			}
#endif /* DEBUG */
			mp->pmap->stats.resident_count--;

			/* release mapping structure and PTE, keeping
			 * track of changed and referenced bits
			 */
			pmap_free_mapping(mp);

			/* Replace the pte entry by the new one */

			match.bits.valid = FALSE;
			pte->pte0 = match;

			return pte;
		}
	}
	return PTE_NULL;
}


/*
 * Collect un-wired mappings. How best to do this??
 */
/* static */
void
pmap_reap_mappings(void)
{
#if DEBUG
	if (pmdebug & PDB_MAPPING)
		printf("pmap_reap_mappings()\n");
#endif

	panic("out of mapping structs");
}

#if	MACH_VM_DEBUG
int
pmap_list_resident_pages(
	register pmap_t		pmap,
	register vm_offset_t	*listp,
	register int		space)
{
	return 0;
}
#endif	/* MACH_VM_DEBUG */

void
pmap_copy_part_page(
	vm_offset_t	src,
	vm_offset_t	src_offset,
	vm_offset_t	dst,
	vm_offset_t	dst_offset,
	vm_size_t	len)
{
	register struct phys_entry *pp_src, *pp_dst;
	int	s;

        assert(((dst & PAGE_MASK)+dst_offset+len) <= PAGE_SIZE);
        assert(((src & PAGE_MASK)+src_offset+len) <= PAGE_SIZE);

#ifdef	TLB_ZERO
	/* for as-yet unused TLB_ZERO - lazily zero the page if needed */
	pp_dst = pmap_find_physentry(dst);
	if (pp_dst != PHYS_NULL) {
		if (pp_dst->tlbprot & TLB_ZERO) {
			phys_bzero(trunc_page(dst), PAGE_SIZE);
			pp_dst->tlbprot &= ~TLB_ZERO;
		}
	}
#endif

	/*
	 * Since the source and destination are physical addresses, 
	 * turn off data translation to perform a  bcopy() in bcopy_phys().
	 */
	phys_copy((vm_offset_t) src+src_offset,
		  (vm_offset_t) dst+dst_offset, len);
}

void
pmap_zero_part_page(
	vm_offset_t	p,
	vm_offset_t     offset,
	vm_size_t       len)
{
    panic("pmap_zero_part_page");
}

/*
 *	kvtophys(addr)
 *
 *	Convert a kernel virtual address to a physical address
 */
vm_offset_t
kvtophys(
	vm_offset_t va)
{
	pte_t *pte;
	
	pte = find_or_allocate_pte(PPC_SID_KERNEL, trunc_page(va), FALSE);
	if (pte == NULL)
		return(0);
	return trunc_page(pte->pte1.word) | (va & page_mask);
}

/*
 *	phystokv(addr)
 *
 *	Convert a physical address to a kernel virtual address if
 *	there is a mapping, otherwise return NULL
 */
vm_offset_t
phystokv(
	vm_offset_t pa)
{
	struct phys_entry *pp;
	struct mapping    *mp;

	pp = pmap_find_physentry(pa);
	if (pp != PHYS_NULL) {
		queue_iterate(&pp->phys_link, mp,
			      struct mapping *, phys_link) {
			if (mp->space == PPC_SID_KERNEL)
				return mp->vm_info.bits.page << PPC_PGSHIFT |
					(pa & (PAGE_SIZE-1));
		}
	}
	return (vm_offset_t)NULL;
}

#if DEBUG
/* This function is included to test compiler functionality, it does
 * a simple slide through the bitfields of a pte to check
 * minimally correct bitfield behaviour. We need bitfields to work!
 */
#define TEST(PTE0,PTE1)							\
	if ((pte.pte0.word != (unsigned int)PTE0) ||			\
	    (pte.pte1.word != (unsigned int)PTE1)) {			\
		printf("BITFIELD TEST FAILED AT LINE %d\n",__LINE__);	\
		printf("EXPECTING 0x%08x and 0x%08x\n",			\
		       (unsigned int)PTE0,(unsigned int)PTE1);		\
		printf("GOT       0x%08x and 0x%08x\n",			\
		       pte.pte0.word,pte.pte1.word);			\
		return 1;						\
	}

extern int pmap_test_bitfields(void); /* Prototyped also in go.c */
int pmap_test_bitfields(void)
{
	pte_t	pte;
	pte.pte0.word = 0;
	pte.pte1.word = 0;

	pte.pte0.bits.segment_id = 0x103; TEST(0x103 << 7, 0);
	pte.pte0.bits.valid = 1;
	TEST((unsigned int)(0x80000000U| 0x103<<7), 0);
	pte.pte0.bits.hash_id = 1;
	TEST((unsigned int)(0x80000000U | 0x103 << 7 | 1<<6), 0);
	pte.pte0.bits.page_index = 3;
	TEST((unsigned int)(0x80000000U|0x103<<7|1<<6|3), 0);

	pte.pte0.bits.segment_id = 0;
	TEST((unsigned int)(0x80000000U|1<<6|3), 0);
	pte.pte0.bits.page_index = 0;
	TEST((unsigned int)(0x80000000U|1<<6), 0);
	pte.pte0.bits.valid = 0;
	TEST(1<<6, 0);

	pte.pte1.bits.referenced = 1; TEST(1<<6, 1<<8);
	pte.pte1.bits.protection = 3; TEST(1<<6, (unsigned int)(1<<8 | 3));
	pte.pte1.bits.changed = 1; TEST(1<<6,(unsigned int)(1<<8|1<<7|3));
	pte.pte1.bits.phys_page = 0xfffff;
	TEST(1<<6,(unsigned int)(0xfffff000U|1<<8|1<<7|3));

	pte.pte1.bits.changed = 0;
	TEST(1<<6,(unsigned int)(0xfffff000U|1<<8|3));

	return 0;
}
#endif /* DEBUG */
