/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux_DR2.1_Update1
 */

/* Things that don't need to be exported from pmap. Putting
 * them here and not in pmap.h avoids major recompiles when
 * modifying something either here or in proc_reg.h
 */

#ifndef _PMAP_INTERNALS_H_
#define _PMAP_INTERNALS_H_

#include <mach/vm_types.h>
#include <mach/machine/vm_types.h>
#include <mach/vm_prot.h>
#include <mach/vm_statistics.h>
#include <kern/queue.h>
#include <ppc/proc_reg.h>

/* Page table entries are stored in groups (PTEGS) in a hash table */

#if __PPC__
#if _BIG_ENDIAN == 0
error - bitfield structures are not checked for bit ordering in words
#endif /* _BIG_ENDIAN */
#endif /* __PPC__ */

/*
 * Don't change these structures unless you change the assembly code in
 * locore.s
 */
struct mapping {
	queue_head_t	phys_link;	/* for mappings of a given PA */
	space_t		space;		/* virtual space */

	union {
		unsigned int word;
		struct {
			vm_offset_t	page:28;/* virtual page number */
			unsigned int    wired:1;  /* boolean */
			unsigned int	:0;
		} bits;
	} vm_info;

	pte_t	        *pte;		/* pointer to pte in hash table */
	struct pmap	*pmap;		/* pmap mapping belongs to */
};

#define MAPPING_NULL	((struct mapping *) 0)

struct phys_entry {
	queue_head_t	phys_link;	/* head of mappings of a given PA */
	pte1_t		pte1;		/* referenced/changed/wimg info */
};

#define PHYS_NULL	((struct phys_entry *)0)

/* Memory may be non-contiguous. This data structure contains info
 * for mapping this non-contiguous space into the contiguous
 * physical->virtual mapping tables. An array of this type is
 * provided to the pmap system at bootstrap by ppc_vm_init.
 *
 * NB : regions must be in order in this structure.
 */

typedef struct mem_region {
	vm_offset_t start;	/* Address of base of region */
	struct phys_entry *phys_table; /* base of region's table */
	unsigned int end;       /* End address+1 */
} mem_region_t;

/* MEM_REGION_MAX has a PowerMac dependancy - at least the value of
 * kMaxRAMBanks in ppc/POWERMAC/nkinfo.h
 */
#define MEM_REGION_MAX 26

extern mem_region_t mem_regions[MEM_REGION_MAX];
extern int          mem_regions_count;

/* keep track of free regions of physical memory so that we can offer
 * them up via pmap_next_page later on
 */

#define FREE_REGION_MAX 8
extern mem_region_t free_regions[FREE_REGION_MAX];
extern int          free_regions_count;

/* Prototypes */

struct phys_entry *pmap_find_physentry(vm_offset_t pa);

extern struct mapping *pmap_find_mapping(space_t space, vm_offset_t offset);

extern void pmap_free_mapping(register struct mapping *mp);

extern void pmap_reap_mappings(void);

static struct mapping *pmap_enter_mapping(pmap_t pmap,
					  space_t space,
					  vm_offset_t va,
					  vm_offset_t pa,
					  pte_t *pte,
					  unsigned prot,
					  struct phys_entry *pp);

extern pte_t *pmap_pteg_overflow(pte_t *primary_hash, pte0_t primary_match,
				 pte_t *secondary_hash, pte0_t secondary_key);

#endif /* _PMAP_INTERNALS_H_ */
